# yapf: disable
# isort: skip_file
# ruff: noqa

# Standard imports.
from __future__ import annotations

# ATTENTION: no types in this file should be imported with their unqualified
#            name (e.g. `from typing import Any`). That would cause clashes
#            with user-defined methods that have the same name. Use
#            fully-qualified names (e.g. `typing.Any`) instead.
import asyncio
import dataclasses
import google.protobuf.descriptor
import google.protobuf.json_format
import google.protobuf.message
import grpc
import grpc_status._async as rpc_status_async
from grpc_status import rpc_status as rpc_status_sync
import json
import logging
import os
import sys
import traceback
import uuid
import rbt
import respect.logging  # type: ignore[import]
import typing
import backoff
from abc import abstractmethod
from datetime import datetime, timedelta
from google.protobuf import empty_pb2, timestamp_pb2
from google.rpc import status_pb2
import rbt.aio.call
import rbt.aio.contexts
import rbt.aio.headers
import rbt.aio.idempotency
import rbt.aio.internals.channel_manager
import rbt.aio.internals.middleware
from rbt.aio.internals.middleware import (
    maybe_run_function_twice_to_validate_effects,
)
import rbt.aio.internals.tasks_cache
import rbt.aio.internals.tasks_dispatcher
import rbt.aio.placement
import rbt.aio.servicers
import rbt.aio.state_managers
import rbt.aio.stubs
import rbt.aio.tasks
import rbt.aio.types
from rbt.aio.types import (
    assert_type,
    service_to_state_type,
    state_type_to_service,
)
import rbt.aio.external
from rbt.settings import MAX_ACTOR_ID_LENGTH, MIN_ACTOR_ID_LENGTH, MAX_BEARER_TOKEN_LENGTH
from rbt.v1alpha1 import tasks_pb2, tasks_pb2_grpc
import typing

# User defined or referenced imports.
import rbt.cloud.v1alpha1.auth.auth_pb2
import rbt.cloud.v1alpha1.auth.auth_pb2_grpc
import rbt.v1alpha1.options_pb2
# Additionally re-export all messages and enums from the pb2 module.
from rbt.cloud.v1alpha1.auth.auth_pb2 import (
    AuthenticateRequest,
    AuthenticateResponse,
    CreateRequest,
    CreateResponse,
    InvalidInputError,
    RevokeRequest,
    RevokeResponse,
)

logger = respect.logging.get_logger(__name__)

# When we are running this code within `node` we might want to log
# things differently. For example, if we already have some logging in
# our TypeScript code these logs may be redundant. Or we want to avoid
# printing stack traces as those are Python specific.
NODEJS: bool = os.environ.get(
    rbt.settings.ENVVAR_RESEMBLE_NODEJS, "false"
).lower() == "true"


def MakeLegacyGrpcServiceable(
    # A legacy gRPC servicer type can't be more specific than `type`,
    # because legacy gRPC servicers (as generated by the gRPC `protoc`
    # plugin) do not share any common base class other than `object`.
    servicer_type: type
) -> rbt.aio.servicers.Serviceable:
    raise ValueError(f"Unknown legacy gRPC servicer type '{servicer_type}'")


# For internal calls, we can use a magic token to bypass token verification and
# authorization checks. The token provides no auth information (e.g.,
# `context.auth is None`).
__internal_magic_token__: str = f'internal-{str(uuid.uuid4())}'

class APIKeyServicerMiddleware(rbt.aio.internals.middleware.Middleware):

    def __init__(
        self,
        *,
        servicer: APIKeyServicer,
        application_id: rbt.aio.types.ApplicationId,
        consensus_id: rbt.aio.types.ConsensusId,
        state_manager: rbt.aio.state_managers.StateManager,
        placement_client: rbt.aio.placement.PlacementClient,
        channel_manager: rbt.aio.internals.channel_manager._ChannelManager,
        tasks_cache: rbt.aio.internals.tasks_cache.TasksCache,
        token_verifier: typing.Optional[rbt.aio.auth.token_verifiers.TokenVerifier],
        effect_validation: rbt.aio.contexts.EffectValidation,
        app_internal_api_key_secret: str,
        ready: asyncio.Event,
    ):
        super().__init__(
            application_id=application_id,
            consensus_id=consensus_id,
            service_name=servicer.__service_name__,
            placement_client=placement_client,
            channel_manager=channel_manager,
            effect_validation=effect_validation,
            app_internal_api_key_secret=app_internal_api_key_secret,
        )

        self._servicer = servicer
        self._state_manager = state_manager
        self.tasks_dispatcher = rbt.aio.internals.tasks_dispatcher.TasksDispatcher(
            application_id=application_id,
            dispatch=self.dispatch,
            tasks_cache=tasks_cache,
            ready=ready,
        )

        # Store the type of each method's request so that stored requests can be
        # deserialized into the correct type.
        self.request_type_by_method_name: dict[str, type[google.protobuf.message.Message]] = {
            'Create': rbt.cloud.v1alpha1.auth.auth_pb2.CreateRequest,
            'Revoke': rbt.cloud.v1alpha1.auth.auth_pb2.RevokeRequest,
            'Authenticate': rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest,
        }

        # Create authorizer.
        self._authorizer: typing.Optional[rbt.aio.auth.authorizers.Authorizer] = servicer.authorizer()

        # Create token verifier.
        self._token_verifier: typing.Optional[rbt.aio.auth.token_verifiers.TokenVerifier] = (
            servicer.token_verifier() or token_verifier
        )

        if self._token_verifier is not None and self._authorizer is None:
            raise RuntimeError(
                f"'TokenVerifier' provided but no 'Authorizer' returned for "
                f"'rbt.cloud.v1alpha1.auth.APIKey'"
            )


    def add_to_server(self, server: grpc.aio.Server) -> None:
        rbt.cloud.v1alpha1.auth.auth_pb2_grpc.add_APIKeyInterfaceServicer_to_server(
            self, server
        )

    async def inspect(self, state_ref: rbt.aio.types.StateRef) -> typing.AsyncIterator[google.protobuf.message.Message]:
        """Implementation of `Middleware.inspect()`."""
        context = self.create_context(
            headers=rbt.aio.headers.Headers(
                application_id=self.application_id,
                service_name=self.service_name,
                state_ref=state_ref,
            ),
            context_type=rbt.aio.contexts.ReaderContext,
        )

        async with self._state_manager.streaming_reader_idempotency_key(
            context, self._servicer.__state_type__
        ) as states:
            async for (state, idempotency_key) in states:
                yield state

    async def react_query(
        self,
        grpc_context: grpc.aio.ServicerContext,
        headers: rbt.aio.headers.Headers,
        method: str,
        request_bytes: bytes,
    ) -> typing.AsyncIterator[tuple[typing.Optional[google.protobuf.message.Message], list[uuid.UUID]]]:
        """Returns the response of calling 'method' given a message
        deserialized from the provided 'request_bytes' for each state
        update that creates a different response.

        # The caller (react.py) should have already ensured that this consensus
        # is authoritative for this traffic.
        assert self.placement_client.consensus_for_actor(
            headers.application_id, headers.service_name, headers.state_ref
        ) == self._consensus_id

        NOTE: only unary reader methods are supported."""
        # Need to define these up here since we can only do that once.
        last_response: typing.Optional[google.protobuf.message.Message] = None
        aggregated_idempotency_keys: list[uuid.UUID] = []
        if method == 'Create':
            # Invariant here is that users should not have called this
            # directly but only through code generated React
            # components which should not have been generated except
            # for valid method candidates.
            logger.warning(
                "Got a React query request with an invalid method name: "
                f"Method '{method}' is invalid for servicer APIKey."
                "\n"
                "Do you have a browser tab open for an older version "
                "of this application, or for a different application all together?"
            )
            raise rbt.aio.aborted.SystemAborted(
                rbt.v1alpha1.errors_pb2.InvalidMethod(),
                message=
                    f"Method '{method}' is invalid"
            )
            yield  # Necessary for type checking.
        elif method == 'Revoke':
            # Invariant here is that users should not have called this
            # directly but only through code generated React
            # components which should not have been generated except
            # for valid method candidates.
            logger.warning(
                "Got a React query request with an invalid method name: "
                f"Method '{method}' is invalid for servicer APIKey."
                "\n"
                "Do you have a browser tab open for an older version "
                "of this application, or for a different application all together?"
            )
            raise rbt.aio.aborted.SystemAborted(
                rbt.v1alpha1.errors_pb2.InvalidMethod(),
                message=
                    f"Method '{method}' is invalid"
            )
            yield  # Necessary for type checking.
        elif method == 'Authenticate':

            try:
                context = self.create_context(
                    headers=headers,
                    context_type=rbt.aio.contexts.ReaderContext,
                )

                context.auth = await self._maybe_verify_token(headers)

                request = rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest()
                request.ParseFromString(request_bytes)

                async with self._state_manager.reactively(
                    context, self._servicer.__state_type__
                ) as states:
                    async for (state, idempotency_keys) in states:
                        # We currently authorize for each new `state`
                        # as a change in the state might now raise a
                        # `PermissionDenied`.
                        await self._maybe_authorize(
                            method_name=
                            'rbt.cloud.v1alpha1.auth.APIKey.Authenticate',
                            headers=headers,
                            auth=context.auth,
                            state=state,
                            request=request,
                        )

                        aggregated_idempotency_keys.extend(idempotency_keys)

                        # Note: This does not do any defensive copying currently:
                        # see https://github.com/reboot-dev/respect/issues/2636.
                        @maybe_run_function_twice_to_validate_effects
                        async def run__Authenticate(validating_effects: bool) -> google.protobuf.message.Message:
                            return await self.__Authenticate(
                                context,
                                state,
                                request,
                                validating_effects=validating_effects,
                            )

                        response = await run__Authenticate()

                        if last_response != response:
                            yield (response, aggregated_idempotency_keys)
                            last_response = response
                        else:
                            yield (None, aggregated_idempotency_keys)

                        aggregated_idempotency_keys.clear()
            except rbt.aio.aborted.Aborted as aborted:
                await grpc_context.abort_with_status(
                    rpc_status_sync.to_status(aborted.to_status())
                )
                raise  # Unreachable but necessary for mypy.
        else:
            logger.warning(
                "Got a React query request with an invalid method name: "
                "Method '{method}' is invalid for servicer APIKey."
                "\n"
                "Do you have a browser tab open for an older version "
                "of this application, or for a different application all together?"
            )
            raise rbt.aio.aborted.SystemAborted(
                rbt.v1alpha1.errors_pb2.InvalidMethod(),
                message=
                    f"Method '{method}' not found"
            )
            yield  # Unreachable but necessary for mypy.

    async def react_mutate(
        self,
        headers: rbt.aio.headers.Headers,
        method: str,
        request_bytes: bytes,
    ) -> google.protobuf.message.Message:
        """Returns the response of calling 'method' given a message
        deserialized from the provided 'request_bytes'."""
        if method == 'Create':
            request = rbt.cloud.v1alpha1.auth.auth_pb2.CreateRequest()
            request.ParseFromString(request_bytes)

            # NOTE: we automatically retry mutations that come through
            # React when we get a `grpc.StatusCode.UNAVAILABLE` to
            # match the retry logic we do in the React code generated
            # to handle lack/loss of connectivity.
            #
            # TODO(benh): revisit this decision if we ever see reason
            # to call `react_mutate()` from any place other than where
            # we're executing React (e.g., browser, next.js server
            # component, etc).
            call_backoff = backoff.Backoff()
            while True:
                # We make a full-fledged gRPC call, so that if this traffic
                # was misrouted (i.e. this consensus is not authoritative
                # for the state), it will now go to the right place. The
                # receiving middleware will handle things like effect
                # validation and so forth.
                assert headers.application_id is not None  # Guaranteed by `Headers`.
                stub = rbt.cloud.v1alpha1.auth.auth_pb2_grpc.APIKeyInterfaceStub(
                    self.channel_manager.get_channel_to(
                        self.placement_client.address_for_actor(
                            headers.application_id,
                            headers.service_name,
                            headers.state_ref,
                        )
                    )
                )
                call = stub.Create(
                    request=request,
                    metadata=headers.to_grpc_metadata(),
                )
                try:
                    return await call
                except grpc.aio.AioRpcError as error:
                    if error.code() == grpc.StatusCode.UNAVAILABLE:
                        await call_backoff()
                        continue

                    # Reconstitute the error that the server threw, if it was a declared error.
                    status = await rpc_status_async.from_call(call)
                    if status is not None:
                        raise APIKey.CreateAborted.from_status(
                            status
                        ) from None
                    raise APIKey.CreateAborted.from_grpc_aio_rpc_error(
                        error
                     ) from None

        elif method == 'Revoke':
            request = rbt.cloud.v1alpha1.auth.auth_pb2.RevokeRequest()
            request.ParseFromString(request_bytes)

            # NOTE: we automatically retry mutations that come through
            # React when we get a `grpc.StatusCode.UNAVAILABLE` to
            # match the retry logic we do in the React code generated
            # to handle lack/loss of connectivity.
            #
            # TODO(benh): revisit this decision if we ever see reason
            # to call `react_mutate()` from any place other than where
            # we're executing React (e.g., browser, next.js server
            # component, etc).
            call_backoff = backoff.Backoff()
            while True:
                # We make a full-fledged gRPC call, so that if this traffic
                # was misrouted (i.e. this consensus is not authoritative
                # for the state), it will now go to the right place. The
                # receiving middleware will handle things like effect
                # validation and so forth.
                assert headers.application_id is not None  # Guaranteed by `Headers`.
                stub = rbt.cloud.v1alpha1.auth.auth_pb2_grpc.APIKeyInterfaceStub(
                    self.channel_manager.get_channel_to(
                        self.placement_client.address_for_actor(
                            headers.application_id,
                            headers.service_name,
                            headers.state_ref,
                        )
                    )
                )
                call = stub.Revoke(
                    request=request,
                    metadata=headers.to_grpc_metadata(),
                )
                try:
                    return await call
                except grpc.aio.AioRpcError as error:
                    if error.code() == grpc.StatusCode.UNAVAILABLE:
                        await call_backoff()
                        continue

                    # Reconstitute the error that the server threw, if it was a declared error.
                    status = await rpc_status_async.from_call(call)
                    if status is not None:
                        raise APIKey.RevokeAborted.from_status(
                            status
                        ) from None
                    raise APIKey.RevokeAborted.from_grpc_aio_rpc_error(
                        error
                     ) from None

        elif method == 'Authenticate':
            # Invariant here is that users should not have called this
            # directly but only through code generated React
            # components which should not have been generated except
            # for valid method candidates.
            logger.warning(
                "Got a react mutate request with an invalid method name: "
                "Method 'Authenticate' is invalid for servicer APIKey."
                "\n"
                "Do you have an old browser tab still open for an older version "
                "of this application, or a different application all together?"
            )
            raise rbt.aio.aborted.SystemAborted(
                rbt.v1alpha1.errors_pb2.InvalidMethod(),
                message=f"Method '{method}' is invalid"
            )
        else:
            logger.warning(
                "Got a react mutate request with an invalid method name: "
                "Method '{method}' is invalid for servicer APIKey."
                "\n"
                "Do you have an old browser tab still open for an older version "
                "of this application, or a different application all together?"
            )
            raise rbt.aio.aborted.SystemAborted(
                rbt.v1alpha1.errors_pb2.InvalidMethod(),
                message=
                    f"Method '{method}' not found"
            )

    async def dispatch(
        self,
        task: rbt.aio.tasks.TaskEffect,
        *,
        only_validate: bool = False,
    ) -> rbt.aio.internals.tasks_dispatcher.TaskResponseOrError:
        """Dispatches the tasks to execute unless 'only_validate' is set to
        true, in which case just ensures that the task actually exists.
        Note that this function will be called *by* tasks_dispatcher; it will
        not itself call into tasks_dispatcher."""

        if 'Create' == task.method_name:
            if only_validate:
                # TODO(benh): validate 'task.request' is correct type.
                return rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse()

            # Use an inline method to create a new scope, so that we can use
            # variable names like `context` and `effects` in multiple branches
            # in this code (notably when there are multiple task types) without
            # hitting a mypy error that the variable's type is not consistent.
            @maybe_run_function_twice_to_validate_effects
            async def run_Create(
                validating_effects: bool,
                context: rbt.aio.contexts.WorkflowContext,
            ):
                async with self._state_manager.task_workflow(
                    context,
                ) as complete:
                    self.tasks_dispatcher.set_complete_task(complete)

                    try:
                        response = await (APIKeyWorkflowStub(
                            context=context,
                            state_ref=context._state_ref,
                        ).Create(
                            typing.cast(rbt.cloud.v1alpha1.auth.auth_pb2.CreateRequest, task.request),
                            bearer_token=__internal_magic_token__,
                            idempotency=rbt.aio.idempotency.Idempotency(
                                alias=f'Task {task.task_id.task_uuid}',
                            ),
                        ))

                        await complete(task, (response, None))
                        return (response, None)

                    except asyncio.CancelledError:
                        # Check if the task was cancelled by a TasksServicer.
                        if self.tasks_dispatcher.is_task_cancelled(task.task_id.task_uuid):
                            # The running task was cancelled by a TasksServicer.
                            await complete(task, (None, tasks_pb2.TaskCancelledError()))
                            return (None, tasks_pb2.TaskCancelledError())
                        else:
                            raise



            return await run_Create(
                self.create_context(
                    headers=rbt.aio.headers.Headers(
                        application_id=self.application_id,
                        service_name=self.service_name,
                        state_ref=rbt.aio.types.StateRef(task.task_id.state_ref),
                    ),
                    context_type=rbt.aio.contexts.WorkflowContext,
                    task=task,
                )
            )
        elif 'Revoke' == task.method_name:
            if only_validate:
                # TODO(benh): validate 'task.request' is correct type.
                return rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse()

            # Use an inline method to create a new scope, so that we can use
            # variable names like `context` and `effects` in multiple branches
            # in this code (notably when there are multiple task types) without
            # hitting a mypy error that the variable's type is not consistent.
            @maybe_run_function_twice_to_validate_effects
            async def run_Revoke(
                validating_effects: bool,
                context: rbt.aio.contexts.WorkflowContext,
            ):
                async with self._state_manager.task_workflow(
                    context,
                ) as complete:
                    self.tasks_dispatcher.set_complete_task(complete)

                    try:
                        response = await (APIKeyWorkflowStub(
                            context=context,
                            state_ref=context._state_ref,
                        ).Revoke(
                            typing.cast(rbt.cloud.v1alpha1.auth.auth_pb2.RevokeRequest, task.request),
                            bearer_token=__internal_magic_token__,
                            idempotency=rbt.aio.idempotency.Idempotency(
                                alias=f'Task {task.task_id.task_uuid}',
                            ),
                        ))

                        await complete(task, (response, None))
                        return (response, None)

                    except asyncio.CancelledError:
                        # Check if the task was cancelled by a TasksServicer.
                        if self.tasks_dispatcher.is_task_cancelled(task.task_id.task_uuid):
                            # The running task was cancelled by a TasksServicer.
                            await complete(task, (None, tasks_pb2.TaskCancelledError()))
                            return (None, tasks_pb2.TaskCancelledError())
                        else:
                            raise



            return await run_Revoke(
                self.create_context(
                    headers=rbt.aio.headers.Headers(
                        application_id=self.application_id,
                        service_name=self.service_name,
                        state_ref=rbt.aio.types.StateRef(task.task_id.state_ref),
                    ),
                    context_type=rbt.aio.contexts.WorkflowContext,
                    task=task,
                )
            )
        elif 'Authenticate' == task.method_name:
            if only_validate:
                # TODO(benh): validate 'task.request' is correct type.
                return rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse()

            # Use an inline method to create a new scope, so that we can use
            # variable names like `context` and `effects` in multiple branches
            # in this code (notably when there are multiple task types) without
            # hitting a mypy error that the variable's type is not consistent.
            @maybe_run_function_twice_to_validate_effects
            async def run_Authenticate(
                validating_effects: bool,
                context: rbt.aio.contexts.WorkflowContext,
            ):
                async with self._state_manager.task_workflow(
                    context,
                ) as complete:
                    self.tasks_dispatcher.set_complete_task(complete)

                    try:
                        response = await (APIKeyWorkflowStub(
                            context=context,
                            state_ref=context._state_ref,
                        ).Authenticate(
                            typing.cast(rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest, task.request),
                            bearer_token=__internal_magic_token__,
                        ))

                        await complete(task, (response, None))
                        return (response, None)

                    except asyncio.CancelledError:
                        # Check if the task was cancelled by a TasksServicer.
                        if self.tasks_dispatcher.is_task_cancelled(task.task_id.task_uuid):
                            # The running task was cancelled by a TasksServicer.
                            await complete(task, (None, tasks_pb2.TaskCancelledError()))
                            return (None, tasks_pb2.TaskCancelledError())
                        else:
                            raise



            return await run_Authenticate(
                self.create_context(
                    headers=rbt.aio.headers.Headers(
                        application_id=self.application_id,
                        service_name=self.service_name,
                        state_ref=rbt.aio.types.StateRef(task.task_id.state_ref),
                    ),
                    context_type=rbt.aio.contexts.WorkflowContext,
                    task=task,
                )
            )

        # There are no tasks for this service.
        start_or_validate = "start" if not only_validate else "validate"
        raise RuntimeError(
            f"Attempted to {start_or_validate} task '{task.method_name}' "
            f"on 'APIKey' which does not exist"
        )

    # APIKey specific methods:
    async def __Create(
        self,
        context: rbt.aio.contexts.WriterContext,
        state: rbt.cloud.v1alpha1.auth.auth_pb2.APIKey,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.CreateRequest,
        *,
        validating_effects: bool,
    ) -> APIKey.CreateEffects:
        try:
            response = (
                await self._servicer.Create(
                    context=context,
                    state=state,
                    request=request
                )
            )
            assert_type(
                response,
                [
                    rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse,
                ],
            )
            # TODO: retrying a control loop might cause an effect
            # validation retry to either "hang" or "do more
            # work". It's not clear that this is always a bug, i.e.,
            # the intent of effect validation for a `workflow` method
            # is to ensure it can handle a failure and the developer
            # might have written the code to handle failures and still
            # run into the "hang" or "do more work" due to effect
            # validation. So for now, we don't bother doing effect
            # validation for loops since we'll be re-running it again.
            if not isinstance(response, rbt.aio.tasks.Loop):
                self.maybe_raise_effect_validation_retry(
                    logger=logger,
                    idempotency_manager=context,
                    method_name='APIKey.Create',
                    validating_effects=validating_effects,
                    context=context,
                )
            return APIKey.CreateEffects(
                state=state,
                response=response,
                tasks=context._tasks,
                _colocated_upserts=context._colocated_upserts,
            )
        except rbt.aio.contexts.RetryReactively:
            # Retrying reactively, just let this propagate.
            raise
        except rbt.aio.contexts.EffectValidationRetry:
            # Doing effect validation, just let this propagate.
            raise
        except rbt.aio.aborted.Aborted as aborted:
            # Log any _unhandled_ abort stack traces to make it
            # easier for debugging.
            #
            # NOTE: we don't log if we're a task as it will be logged
            # in `rbt/aio/internals/tasks_dispatcher.py` instead.
            aborted_type: typing.Optional[type] = None
            aborted_type = APIKey.CreateAborted
            if isinstance(aborted, rbt.aio.aborted.SystemAborted):
                # Not logging when within `node` as we already log there.
                if not NODEJS:
                    logger.warning(
                        f"Unhandled (in 'rbt.cloud.v1alpha1.auth.APIKey.Create') {aborted}; propagating as 'Unknown'\n" +
                        ''.join(traceback.format_exception(aborted))
                    )
                raise rbt.aio.aborted.SystemAborted(
                    rbt.v1alpha1.errors_pb2.Unknown(),
                    # TODO(benh): consider whether or not we want to
                    # include the 'package.service.method' which may
                    # get concatenated together forming a kind of
                    # "stack trace"; while it's super helpful for
                    # debugging, it does expose implementation
                    # information.
                    message=f"unhandled (in 'rbt.cloud.v1alpha1.auth.APIKey.Create') {aborted}"
                )
            else:
                if (
                    aborted_type is not None and
                    not isinstance(aborted, aborted_type) and
                    aborted_type.is_declared_error(aborted.error)
                ):
                    if context.task is None:
                        logger.warning(
                            f"Propagating unhandled (in 'rbt.cloud.v1alpha1.auth.APIKey.Create') {aborted}"
                        )
                elif (
                    aborted_type is None or
                    not isinstance(aborted, aborted_type)
                ):
                    # Not logging when within `node` as we already log there.
                    if not NODEJS:
                        logger.warning(
                            f"Unhandled (in 'rbt.cloud.v1alpha1.auth.APIKey.Create') {aborted}; propagating as 'Unknown'\n" +
                            ''.join(traceback.format_exception(aborted))
                        )
                    raise rbt.aio.aborted.SystemAborted(
                        rbt.v1alpha1.errors_pb2.Unknown(),
                        # TODO(benh): consider whether or not we want to
                        # include the 'package.service.method' which may
                        # get concatenated together forming a kind of
                        # "stack trace"; while it's super helpful for
                        # debugging, it does expose implementation
                        # information.
                        message=f"unhandled (in 'rbt.cloud.v1alpha1.auth.APIKey.Create') {aborted}"
                    )

            raise
        except asyncio.CancelledError:
            # It's pretty normal for an RPC to be cancelled; it's not useful to
            # print a stack trace.
            raise
        except BaseException as exception:
            # Not logging when within `node` as we already log there.
            if not NODEJS:
                logger.warning(
                    f"Unhandled (in 'rbt.cloud.v1alpha1.auth.APIKey.Create') {type(exception).__name__}: {exception}; propagating as 'Unknown'\n" +
                    ''.join(traceback.format_exception(exception))
                )
            raise rbt.aio.aborted.SystemAborted(
                rbt.v1alpha1.errors_pb2.Unknown(),
                # TODO(benh): consider whether or not we want to
                # include the 'package.service.method' which may
                # get concatenated together forming a kind of
                # "stack trace"; while it's super helpful for
                # debugging, it does expose implementation
                # information.
                message=f"unhandled (in 'rbt.cloud.v1alpha1.auth.APIKey.Create') {type(exception).__name__}: {exception}"
            )

    async def _Create(
        self,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.CreateRequest,
        context: rbt.aio.contexts.WriterContext,
        *,
        validating_effects: bool,
        grpc_context: typing.Optional[grpc.aio.ServicerContext] = None,
    ) -> rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse:
        # Try to verify the token if a token verifier exists.
        context.auth = await self._maybe_verify_token(context._headers)

        # Check if we already have performed this mutation!
        #
        # We do this _before_ calling 'transactionally()' because
        # if this call is for a transaction method _and_ we've
        # already performed the transaction then we don't want to
        # become a transaction participant (again) we just want to
        # return the transaction's response.
        idempotent_mutation = self._state_manager.check_for_idempotent_mutation(
            context
        )

        if idempotent_mutation is not None:
            response = rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse()
            response.ParseFromString(idempotent_mutation.response)
            return response

        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__state_type_name__, context._state_ref
                )
            # TODO: this loads a `state` object even when we know we're calling a constructor,
            # in which case (by definition) there is no state. Can we (and is it worth
            # the effort to) change this call to avoid that overhead in that case?
            async with self._state_manager.writer(
                context,
                self._servicer.__state_type__,
                self.tasks_dispatcher,
                transaction=transaction,
                from_constructor=True,
                requires_constructor=True,
            ) as (state, writer):

                await self._maybe_authorize(
                    method_name=
                    'rbt.cloud.v1alpha1.auth.APIKey.Create',
                    headers=context._headers,
                    auth=context.auth,
                    request=request,
                )

                effects = await self.__Create(
                    context,
                    state,
                    request,
                    validating_effects=validating_effects,
                )

                await writer.complete(effects)

                # TODO: We need a single `Effects` superclass for all methods, so we
                # would need to make it "partially" generic (with per-method subclasses
                # filling out the rest of the generic parameters) in order to fix this.
                return effects.response  # type: ignore[return-value]

    async def _schedule_Create(
        self,
        *,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.CreateRequest,
        headers: rbt.aio.headers.Headers,
        grpc_context: grpc.aio.ServicerContext,
    ) -> tuple[rbt.aio.contexts.WriterContext, rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse]:
        context: rbt.aio.contexts.WriterContext = self.create_context(
            headers=headers,
            context_type=rbt.aio.contexts.WriterContext,
        )
        response = rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse()

        # Check if we already have performed this mutation!
        #
        # We do this _before_ calling 'transactionally()' because
        # if this call is for a transaction method _and_ we've
        # already performed the transaction then we don't want to
        # become a transaction participant (again) we just want to
        # return the transaction's response.
        idempotent_mutation = self._state_manager.check_for_idempotent_mutation(
            context
        )

        if idempotent_mutation is not None:
            response.ParseFromString(idempotent_mutation.response)

            # We should have only scheduled a single task!
            assert len(idempotent_mutation.task_ids) == 1
            assert grpc_context is not None
            grpc_context.set_trailing_metadata(
                grpc_context.trailing_metadata() +
                (
                    (
                        rbt.aio.headers.TASK_ID_UUID,
                        str(uuid.UUID(bytes=idempotent_mutation.task_ids[0].task_uuid))
                    ),
                )
            )

            return context, response

        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__state_type_name__, context._state_ref
                )
            async with self._state_manager.writer(
                context,
                self._servicer.__state_type__,
                self.tasks_dispatcher,
                transaction=transaction,
                from_constructor=True,
                requires_constructor=True
            ) as (state, writer):

                # Try to verify the token if a token verifier exists.
                context.auth = await self._maybe_verify_token(headers)

                # Check if we are allowed schedule the task.
                await self._maybe_authorize(
                    method_name=
                    'rbt.cloud.v1alpha1.auth.APIKey.Create',
                    headers=context._headers,
                    auth=context.auth,
                    request=request,
                )

                task = APIKeyServicerTasksStub(
                    context=context,
                    state_ref=context._state_ref,
                ).Create(
                    request,
                    schedule=context._headers.task_schedule,
                )

                effects = rbt.aio.state_managers.Effects(
                    response=response,
                    state=state,
                    tasks=[task],
                )

                assert effects.tasks is not None

                await writer.complete(effects)

                assert grpc_context is not None

                grpc_context.set_trailing_metadata(
                    grpc_context.trailing_metadata() +
                    (
                        (
                            rbt.aio.headers.TASK_ID_UUID,
                            str(uuid.UUID(bytes=task.task_id.task_uuid))
                        ),
                    )
                )

                return context, response

        return context, response


    # Entrypoint for non-reactive network calls (i.e. typical gRPC calls).
    async def Create(
        self,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.CreateRequest,
        grpc_context: grpc.aio.ServicerContext,
    ) -> rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse:
        headers = rbt.aio.headers.Headers.from_grpc_context(grpc_context)
        assert headers.application_id is not None  # Guaranteed by `Headers`.

        # Confirm whether this is the right consensus to be serving this
        # request.
        authoritative_consensus = self.placement_client.consensus_for_actor(
            headers.application_id, headers.service_name, headers.state_ref
        )
        if authoritative_consensus != self.consensus_id:
            # This is NOT the correct consensus. Forward to the correct one.
            correct_address = self.placement_client.address_for_consensus(authoritative_consensus)
            channel = self.channel_manager.get_channel_to(correct_address)
            stub = rbt.cloud.v1alpha1.auth.auth_pb2_grpc.APIKeyInterfaceStub(channel)
            call = stub.Create(
                request=request,
                metadata=grpc_context.invocation_metadata(),
            )
            response = await call
            # Make sure any trailing metadata on the response is forwarded back.
            grpc_context.set_trailing_metadata(await call.trailing_metadata())
            return response


        @maybe_run_function_twice_to_validate_effects
        async def _run(
            validating_effects: bool,
        ) -> rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse:
            context: typing.Optional[rbt.aio.contexts.Context] = None
            try:
                if headers.task_schedule is not None:
                    context, response = await self._schedule_Create(
                        headers=headers,
                        request=request,
                        grpc_context=grpc_context,
                    )
                    return response

                context = self.create_context(
                    headers=headers,
                    context_type=rbt.aio.contexts.WriterContext,
                )
                assert context is not None

                return await self._Create(
                    request,
                    context,
                    validating_effects=validating_effects,
                    grpc_context=grpc_context,
                )
            except rbt.aio.contexts.EffectValidationRetry:
                # Doing effect validation, just let this propagate.
                raise
            except rbt.aio.aborted.Aborted as aborted:
                await grpc_context.abort_with_status(
                    rpc_status_sync.to_status(aborted.to_status())
                )
                raise  # Unreachable but necessary for mypy.
            except asyncio.CancelledError:
                # It's pretty normal for an RPC to be cancelled; it's not useful to
                # print a stack trace.
                raise
            except BaseException as exception:
                # Print the exception stack trace for easier debugging. Note
                # that we don't include the stack trace in an error message
                # for the same reason that gRPC doesn't do so by default,
                # see https://github.com/grpc/grpc/issues/14897, but since this
                # should only get logged on the server side it is safe.
                logger.warning(
                    'Unhandled exception\n' +
                    ''.join(traceback.format_exc() if not NODEJS else [f"{type(exception).__name__}: {exception}"])
                )

                # Re-raise the exception for gRPC to handle!
                raise
            finally:
                if context is not None and context.transaction_id is not None:
                    # Propagate transaction participants.
                    grpc_context.set_trailing_metadata(
                        grpc_context.trailing_metadata() +
                        context.participants.to_grpc_metadata()
                    )

        return await _run()

    async def __Revoke(
        self,
        context: rbt.aio.contexts.WriterContext,
        state: rbt.cloud.v1alpha1.auth.auth_pb2.APIKey,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.RevokeRequest,
        *,
        validating_effects: bool,
    ) -> APIKey.RevokeEffects:
        try:
            response = (
                await self._servicer.Revoke(
                    context=context,
                    state=state,
                    request=request
                )
            )
            assert_type(
                response,
                [
                    rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse,
                ],
            )
            # TODO: retrying a control loop might cause an effect
            # validation retry to either "hang" or "do more
            # work". It's not clear that this is always a bug, i.e.,
            # the intent of effect validation for a `workflow` method
            # is to ensure it can handle a failure and the developer
            # might have written the code to handle failures and still
            # run into the "hang" or "do more work" due to effect
            # validation. So for now, we don't bother doing effect
            # validation for loops since we'll be re-running it again.
            if not isinstance(response, rbt.aio.tasks.Loop):
                self.maybe_raise_effect_validation_retry(
                    logger=logger,
                    idempotency_manager=context,
                    method_name='APIKey.Revoke',
                    validating_effects=validating_effects,
                    context=context,
                )
            return APIKey.RevokeEffects(
                state=state,
                response=response,
                tasks=context._tasks,
                _colocated_upserts=context._colocated_upserts,
            )
        except rbt.aio.contexts.RetryReactively:
            # Retrying reactively, just let this propagate.
            raise
        except rbt.aio.contexts.EffectValidationRetry:
            # Doing effect validation, just let this propagate.
            raise
        except rbt.aio.aborted.Aborted as aborted:
            # Log any _unhandled_ abort stack traces to make it
            # easier for debugging.
            #
            # NOTE: we don't log if we're a task as it will be logged
            # in `rbt/aio/internals/tasks_dispatcher.py` instead.
            aborted_type: typing.Optional[type] = None
            aborted_type = APIKey.RevokeAborted
            if isinstance(aborted, rbt.aio.aborted.SystemAborted):
                # Not logging when within `node` as we already log there.
                if not NODEJS:
                    logger.warning(
                        f"Unhandled (in 'rbt.cloud.v1alpha1.auth.APIKey.Revoke') {aborted}; propagating as 'Unknown'\n" +
                        ''.join(traceback.format_exception(aborted))
                    )
                raise rbt.aio.aborted.SystemAborted(
                    rbt.v1alpha1.errors_pb2.Unknown(),
                    # TODO(benh): consider whether or not we want to
                    # include the 'package.service.method' which may
                    # get concatenated together forming a kind of
                    # "stack trace"; while it's super helpful for
                    # debugging, it does expose implementation
                    # information.
                    message=f"unhandled (in 'rbt.cloud.v1alpha1.auth.APIKey.Revoke') {aborted}"
                )
            else:
                if (
                    aborted_type is not None and
                    not isinstance(aborted, aborted_type) and
                    aborted_type.is_declared_error(aborted.error)
                ):
                    if context.task is None:
                        logger.warning(
                            f"Propagating unhandled (in 'rbt.cloud.v1alpha1.auth.APIKey.Revoke') {aborted}"
                        )
                elif (
                    aborted_type is None or
                    not isinstance(aborted, aborted_type)
                ):
                    # Not logging when within `node` as we already log there.
                    if not NODEJS:
                        logger.warning(
                            f"Unhandled (in 'rbt.cloud.v1alpha1.auth.APIKey.Revoke') {aborted}; propagating as 'Unknown'\n" +
                            ''.join(traceback.format_exception(aborted))
                        )
                    raise rbt.aio.aborted.SystemAborted(
                        rbt.v1alpha1.errors_pb2.Unknown(),
                        # TODO(benh): consider whether or not we want to
                        # include the 'package.service.method' which may
                        # get concatenated together forming a kind of
                        # "stack trace"; while it's super helpful for
                        # debugging, it does expose implementation
                        # information.
                        message=f"unhandled (in 'rbt.cloud.v1alpha1.auth.APIKey.Revoke') {aborted}"
                    )

            raise
        except asyncio.CancelledError:
            # It's pretty normal for an RPC to be cancelled; it's not useful to
            # print a stack trace.
            raise
        except BaseException as exception:
            # Not logging when within `node` as we already log there.
            if not NODEJS:
                logger.warning(
                    f"Unhandled (in 'rbt.cloud.v1alpha1.auth.APIKey.Revoke') {type(exception).__name__}: {exception}; propagating as 'Unknown'\n" +
                    ''.join(traceback.format_exception(exception))
                )
            raise rbt.aio.aborted.SystemAborted(
                rbt.v1alpha1.errors_pb2.Unknown(),
                # TODO(benh): consider whether or not we want to
                # include the 'package.service.method' which may
                # get concatenated together forming a kind of
                # "stack trace"; while it's super helpful for
                # debugging, it does expose implementation
                # information.
                message=f"unhandled (in 'rbt.cloud.v1alpha1.auth.APIKey.Revoke') {type(exception).__name__}: {exception}"
            )

    async def _Revoke(
        self,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.RevokeRequest,
        context: rbt.aio.contexts.WriterContext,
        *,
        validating_effects: bool,
        grpc_context: typing.Optional[grpc.aio.ServicerContext] = None,
    ) -> rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse:
        # Try to verify the token if a token verifier exists.
        context.auth = await self._maybe_verify_token(context._headers)

        # Check if we already have performed this mutation!
        #
        # We do this _before_ calling 'transactionally()' because
        # if this call is for a transaction method _and_ we've
        # already performed the transaction then we don't want to
        # become a transaction participant (again) we just want to
        # return the transaction's response.
        idempotent_mutation = self._state_manager.check_for_idempotent_mutation(
            context
        )

        if idempotent_mutation is not None:
            response = rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse()
            response.ParseFromString(idempotent_mutation.response)
            return response

        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__state_type_name__, context._state_ref
                )
            # TODO: this loads a `state` object even when we know we're calling a constructor,
            # in which case (by definition) there is no state. Can we (and is it worth
            # the effort to) change this call to avoid that overhead in that case?
            async with self._state_manager.writer(
                context,
                self._servicer.__state_type__,
                self.tasks_dispatcher,
                transaction=transaction,
                from_constructor=False,
                requires_constructor=True,
            ) as (state, writer):

                await self._maybe_authorize(
                    method_name=
                    'rbt.cloud.v1alpha1.auth.APIKey.Revoke',
                    headers=context._headers,
                    auth=context.auth,
                    state=state,
                    request=request,
                )

                effects = await self.__Revoke(
                    context,
                    state,
                    request,
                    validating_effects=validating_effects,
                )

                await writer.complete(effects)

                # TODO: We need a single `Effects` superclass for all methods, so we
                # would need to make it "partially" generic (with per-method subclasses
                # filling out the rest of the generic parameters) in order to fix this.
                return effects.response  # type: ignore[return-value]

    async def _schedule_Revoke(
        self,
        *,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.RevokeRequest,
        headers: rbt.aio.headers.Headers,
        grpc_context: grpc.aio.ServicerContext,
    ) -> tuple[rbt.aio.contexts.WriterContext, rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse]:
        context: rbt.aio.contexts.WriterContext = self.create_context(
            headers=headers,
            context_type=rbt.aio.contexts.WriterContext,
        )
        response = rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse()

        # Check if we already have performed this mutation!
        #
        # We do this _before_ calling 'transactionally()' because
        # if this call is for a transaction method _and_ we've
        # already performed the transaction then we don't want to
        # become a transaction participant (again) we just want to
        # return the transaction's response.
        idempotent_mutation = self._state_manager.check_for_idempotent_mutation(
            context
        )

        if idempotent_mutation is not None:
            response.ParseFromString(idempotent_mutation.response)

            # We should have only scheduled a single task!
            assert len(idempotent_mutation.task_ids) == 1
            assert grpc_context is not None
            grpc_context.set_trailing_metadata(
                grpc_context.trailing_metadata() +
                (
                    (
                        rbt.aio.headers.TASK_ID_UUID,
                        str(uuid.UUID(bytes=idempotent_mutation.task_ids[0].task_uuid))
                    ),
                )
            )

            return context, response

        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__state_type_name__, context._state_ref
                )
            async with self._state_manager.writer(
                context,
                self._servicer.__state_type__,
                self.tasks_dispatcher,
                transaction=transaction,
                from_constructor=False,
                requires_constructor=True
            ) as (state, writer):

                # Try to verify the token if a token verifier exists.
                context.auth = await self._maybe_verify_token(headers)

                # Check if we are allowed schedule the task.
                await self._maybe_authorize(
                    method_name=
                    'rbt.cloud.v1alpha1.auth.APIKey.Revoke',
                    headers=context._headers,
                    auth=context.auth,
                    state=state,
                    request=request,
                )

                task = APIKeyServicerTasksStub(
                    context=context,
                    state_ref=context._state_ref,
                ).Revoke(
                    request,
                    schedule=context._headers.task_schedule,
                )

                effects = rbt.aio.state_managers.Effects(
                    response=response,
                    state=state,
                    tasks=[task],
                )

                assert effects.tasks is not None

                await writer.complete(effects)

                assert grpc_context is not None

                grpc_context.set_trailing_metadata(
                    grpc_context.trailing_metadata() +
                    (
                        (
                            rbt.aio.headers.TASK_ID_UUID,
                            str(uuid.UUID(bytes=task.task_id.task_uuid))
                        ),
                    )
                )

                return context, response

        return context, response


    # Entrypoint for non-reactive network calls (i.e. typical gRPC calls).
    async def Revoke(
        self,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.RevokeRequest,
        grpc_context: grpc.aio.ServicerContext,
    ) -> rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse:
        headers = rbt.aio.headers.Headers.from_grpc_context(grpc_context)
        assert headers.application_id is not None  # Guaranteed by `Headers`.

        # Confirm whether this is the right consensus to be serving this
        # request.
        authoritative_consensus = self.placement_client.consensus_for_actor(
            headers.application_id, headers.service_name, headers.state_ref
        )
        if authoritative_consensus != self.consensus_id:
            # This is NOT the correct consensus. Forward to the correct one.
            correct_address = self.placement_client.address_for_consensus(authoritative_consensus)
            channel = self.channel_manager.get_channel_to(correct_address)
            stub = rbt.cloud.v1alpha1.auth.auth_pb2_grpc.APIKeyInterfaceStub(channel)
            call = stub.Revoke(
                request=request,
                metadata=grpc_context.invocation_metadata(),
            )
            response = await call
            # Make sure any trailing metadata on the response is forwarded back.
            grpc_context.set_trailing_metadata(await call.trailing_metadata())
            return response


        @maybe_run_function_twice_to_validate_effects
        async def _run(
            validating_effects: bool,
        ) -> rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse:
            context: typing.Optional[rbt.aio.contexts.Context] = None
            try:
                if headers.task_schedule is not None:
                    context, response = await self._schedule_Revoke(
                        headers=headers,
                        request=request,
                        grpc_context=grpc_context,
                    )
                    return response

                context = self.create_context(
                    headers=headers,
                    context_type=rbt.aio.contexts.WriterContext,
                )
                assert context is not None

                return await self._Revoke(
                    request,
                    context,
                    validating_effects=validating_effects,
                    grpc_context=grpc_context,
                )
            except rbt.aio.contexts.EffectValidationRetry:
                # Doing effect validation, just let this propagate.
                raise
            except rbt.aio.aborted.Aborted as aborted:
                await grpc_context.abort_with_status(
                    rpc_status_sync.to_status(aborted.to_status())
                )
                raise  # Unreachable but necessary for mypy.
            except asyncio.CancelledError:
                # It's pretty normal for an RPC to be cancelled; it's not useful to
                # print a stack trace.
                raise
            except BaseException as exception:
                # Print the exception stack trace for easier debugging. Note
                # that we don't include the stack trace in an error message
                # for the same reason that gRPC doesn't do so by default,
                # see https://github.com/grpc/grpc/issues/14897, but since this
                # should only get logged on the server side it is safe.
                logger.warning(
                    'Unhandled exception\n' +
                    ''.join(traceback.format_exc() if not NODEJS else [f"{type(exception).__name__}: {exception}"])
                )

                # Re-raise the exception for gRPC to handle!
                raise
            finally:
                if context is not None and context.transaction_id is not None:
                    # Propagate transaction participants.
                    grpc_context.set_trailing_metadata(
                        grpc_context.trailing_metadata() +
                        context.participants.to_grpc_metadata()
                    )

        return await _run()

    async def __Authenticate(
        self,
        context: rbt.aio.contexts.ReaderContext,
        state: rbt.cloud.v1alpha1.auth.auth_pb2.APIKey,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest,
        *,
        validating_effects: bool,
    ) -> rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse:
        try:
            response = (
                await self._servicer.Authenticate(
                    context=context,
                    state=state,
                    request=request
                )
            )
            assert_type(
                response,
                [
                    rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse,
                ],
            )
            # TODO: retrying a control loop might cause an effect
            # validation retry to either "hang" or "do more
            # work". It's not clear that this is always a bug, i.e.,
            # the intent of effect validation for a `workflow` method
            # is to ensure it can handle a failure and the developer
            # might have written the code to handle failures and still
            # run into the "hang" or "do more work" due to effect
            # validation. So for now, we don't bother doing effect
            # validation for loops since we'll be re-running it again.
            if not isinstance(response, rbt.aio.tasks.Loop):
                self.maybe_raise_effect_validation_retry(
                    logger=logger,
                    idempotency_manager=context,
                    method_name='APIKey.Authenticate',
                    validating_effects=validating_effects,
                    context=context,
                )
            return response
        except rbt.aio.contexts.RetryReactively:
            # Retrying reactively, just let this propagate.
            raise
        except rbt.aio.contexts.EffectValidationRetry:
            # Doing effect validation, just let this propagate.
            raise
        except rbt.aio.aborted.Aborted as aborted:
            # Log any _unhandled_ abort stack traces to make it
            # easier for debugging.
            #
            # NOTE: we don't log if we're a task as it will be logged
            # in `rbt/aio/internals/tasks_dispatcher.py` instead.
            aborted_type: typing.Optional[type] = None
            aborted_type = APIKey.AuthenticateAborted
            if isinstance(aborted, rbt.aio.aborted.SystemAborted):
                # Not logging when within `node` as we already log there.
                if not NODEJS:
                    logger.warning(
                        f"Unhandled (in 'rbt.cloud.v1alpha1.auth.APIKey.Authenticate') {aborted}; propagating as 'Unknown'\n" +
                        ''.join(traceback.format_exception(aborted))
                    )
                raise rbt.aio.aborted.SystemAborted(
                    rbt.v1alpha1.errors_pb2.Unknown(),
                    # TODO(benh): consider whether or not we want to
                    # include the 'package.service.method' which may
                    # get concatenated together forming a kind of
                    # "stack trace"; while it's super helpful for
                    # debugging, it does expose implementation
                    # information.
                    message=f"unhandled (in 'rbt.cloud.v1alpha1.auth.APIKey.Authenticate') {aborted}"
                )
            else:
                if (
                    aborted_type is not None and
                    not isinstance(aborted, aborted_type) and
                    aborted_type.is_declared_error(aborted.error)
                ):
                    if context.task is None:
                        logger.warning(
                            f"Propagating unhandled (in 'rbt.cloud.v1alpha1.auth.APIKey.Authenticate') {aborted}"
                        )
                elif (
                    aborted_type is None or
                    not isinstance(aborted, aborted_type)
                ):
                    # Not logging when within `node` as we already log there.
                    if not NODEJS:
                        logger.warning(
                            f"Unhandled (in 'rbt.cloud.v1alpha1.auth.APIKey.Authenticate') {aborted}; propagating as 'Unknown'\n" +
                            ''.join(traceback.format_exception(aborted))
                        )
                    raise rbt.aio.aborted.SystemAborted(
                        rbt.v1alpha1.errors_pb2.Unknown(),
                        # TODO(benh): consider whether or not we want to
                        # include the 'package.service.method' which may
                        # get concatenated together forming a kind of
                        # "stack trace"; while it's super helpful for
                        # debugging, it does expose implementation
                        # information.
                        message=f"unhandled (in 'rbt.cloud.v1alpha1.auth.APIKey.Authenticate') {aborted}"
                    )

            raise
        except asyncio.CancelledError:
            # It's pretty normal for an RPC to be cancelled; it's not useful to
            # print a stack trace.
            raise
        except BaseException as exception:
            # Not logging when within `node` as we already log there.
            if not NODEJS:
                logger.warning(
                    f"Unhandled (in 'rbt.cloud.v1alpha1.auth.APIKey.Authenticate') {type(exception).__name__}: {exception}; propagating as 'Unknown'\n" +
                    ''.join(traceback.format_exception(exception))
                )
            raise rbt.aio.aborted.SystemAborted(
                rbt.v1alpha1.errors_pb2.Unknown(),
                # TODO(benh): consider whether or not we want to
                # include the 'package.service.method' which may
                # get concatenated together forming a kind of
                # "stack trace"; while it's super helpful for
                # debugging, it does expose implementation
                # information.
                message=f"unhandled (in 'rbt.cloud.v1alpha1.auth.APIKey.Authenticate') {type(exception).__name__}: {exception}"
            )

    async def _Authenticate(
        self,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest,
        context: rbt.aio.contexts.ReaderContext,
        *,
        validating_effects: bool,
        grpc_context: typing.Optional[grpc.aio.ServicerContext] = None,
    ) -> rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse:
        # Try to verify the token if a token verifier exists.
        context.auth = await self._maybe_verify_token(context._headers)


        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__state_type_name__, context._state_ref
                )
            async with self._state_manager.reader(
                context, self._servicer.__state_type__
            ) as state:
                await self._maybe_authorize(
                    method_name=
                    'rbt.cloud.v1alpha1.auth.APIKey.Authenticate',
                    headers=context._headers,
                    auth=context.auth,
                    state=state,
                    request=request,
                )
                response = await self.__Authenticate(
                    context,
                    state,
                    request,
                    validating_effects=validating_effects,
                )
                if transaction is not None:
                    # We need the transaction to be stored _before_
                    # returning a response to the user otherwise they
                    # may read two different states if we crash.
                    await self._state_manager.transaction_participant_store(
                        transaction
                    )
                return response

    async def _schedule_Authenticate(
        self,
        *,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest,
        headers: rbt.aio.headers.Headers,
        grpc_context: grpc.aio.ServicerContext,
    ) -> tuple[rbt.aio.contexts.WriterContext, rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse]:
        context: rbt.aio.contexts.WriterContext = self.create_context(
            headers=headers,
            context_type=rbt.aio.contexts.WriterContext,
        )
        response = rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse()


        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__state_type_name__, context._state_ref
                )
            async with self._state_manager.writer(
                context,
                self._servicer.__state_type__,
                self.tasks_dispatcher,
                transaction=transaction,
                from_constructor=False,
                requires_constructor=True
            ) as (state, writer):

                # Try to verify the token if a token verifier exists.
                context.auth = await self._maybe_verify_token(headers)

                # Check if we are allowed schedule the task.
                await self._maybe_authorize(
                    method_name=
                    'rbt.cloud.v1alpha1.auth.APIKey.Authenticate',
                    headers=context._headers,
                    auth=context.auth,
                    state=state,
                    request=request,
                )

                task = APIKeyServicerTasksStub(
                    context=context,
                    state_ref=context._state_ref,
                ).Authenticate(
                    request,
                    schedule=context._headers.task_schedule,
                )

                effects = rbt.aio.state_managers.Effects(
                    response=response,
                    state=state,
                    tasks=[task],
                )

                assert effects.tasks is not None

                await writer.complete(effects)

                assert grpc_context is not None

                grpc_context.set_trailing_metadata(
                    grpc_context.trailing_metadata() +
                    (
                        (
                            rbt.aio.headers.TASK_ID_UUID,
                            str(uuid.UUID(bytes=task.task_id.task_uuid))
                        ),
                    )
                )

                return context, response

        return context, response


    # Entrypoint for non-reactive network calls (i.e. typical gRPC calls).
    async def Authenticate(
        self,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest,
        grpc_context: grpc.aio.ServicerContext,
    ) -> rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse:
        headers = rbt.aio.headers.Headers.from_grpc_context(grpc_context)
        assert headers.application_id is not None  # Guaranteed by `Headers`.

        # Confirm whether this is the right consensus to be serving this
        # request.
        authoritative_consensus = self.placement_client.consensus_for_actor(
            headers.application_id, headers.service_name, headers.state_ref
        )
        if authoritative_consensus != self.consensus_id:
            # This is NOT the correct consensus. Forward to the correct one.
            correct_address = self.placement_client.address_for_consensus(authoritative_consensus)
            channel = self.channel_manager.get_channel_to(correct_address)
            stub = rbt.cloud.v1alpha1.auth.auth_pb2_grpc.APIKeyInterfaceStub(channel)
            call = stub.Authenticate(
                request=request,
                metadata=grpc_context.invocation_metadata(),
            )
            response = await call
            # Make sure any trailing metadata on the response is forwarded back.
            grpc_context.set_trailing_metadata(await call.trailing_metadata())
            return response


        @maybe_run_function_twice_to_validate_effects
        async def _run(
            validating_effects: bool,
        ) -> rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse:
            context: typing.Optional[rbt.aio.contexts.Context] = None
            try:
                if headers.task_schedule is not None:
                    context, response = await self._schedule_Authenticate(
                        headers=headers,
                        request=request,
                        grpc_context=grpc_context,
                    )
                    return response

                context = self.create_context(
                    headers=headers,
                    context_type=rbt.aio.contexts.ReaderContext,
                )
                assert context is not None

                return await self._Authenticate(
                    request,
                    context,
                    validating_effects=validating_effects,
                    grpc_context=grpc_context,
                )
            except rbt.aio.contexts.EffectValidationRetry:
                # Doing effect validation, just let this propagate.
                raise
            except rbt.aio.aborted.Aborted as aborted:
                await grpc_context.abort_with_status(
                    rpc_status_sync.to_status(aborted.to_status())
                )
                raise  # Unreachable but necessary for mypy.
            except asyncio.CancelledError:
                # It's pretty normal for an RPC to be cancelled; it's not useful to
                # print a stack trace.
                raise
            except BaseException as exception:
                # Print the exception stack trace for easier debugging. Note
                # that we don't include the stack trace in an error message
                # for the same reason that gRPC doesn't do so by default,
                # see https://github.com/grpc/grpc/issues/14897, but since this
                # should only get logged on the server side it is safe.
                logger.warning(
                    'Unhandled exception\n' +
                    ''.join(traceback.format_exc() if not NODEJS else [f"{type(exception).__name__}: {exception}"])
                )

                # Re-raise the exception for gRPC to handle!
                raise
            finally:
                if context is not None and context.transaction_id is not None:
                    # Propagate transaction participants.
                    grpc_context.set_trailing_metadata(
                        grpc_context.trailing_metadata() +
                        context.participants.to_grpc_metadata()
                    )

        return await _run()

    async def _maybe_authorize(
        self,
        *,
        method_name: str,
        headers: rbt.aio.headers.Headers,
        auth: typing.Optional[rbt.aio.auth.Auth],
        state: typing.Optional[APIKeyStateType] = None,
        request: typing.Optional[APIKeyRequestTypes] = None,
    ) -> None:
        """Check authorization decision for the given method, using available
        information.

        Raises `PermissionDenied` in case authorizer is present but the request
        is not authorized.
        """
        # To authorize internal calls, we use an internal magic token.
        if headers.bearer_token == __internal_magic_token__:
            return

        if self._authorizer is None:
            return

        if not self._authorizer.can_authorize(method_name):
            # Better safe than sorry.
            raise rbt.aio.aborted.SystemAborted(
                rbt.v1alpha1.errors_pb2.PermissionDenied(),
                message=
                    f"Authorizer does not support method '{method_name}'"
            )

        # If there is state, ensure that we don't accidentally mutate it.
        if state is not None:
            # TODO: If we can read-only a state object we can avoid a copy here.
            state_copy = APIKeyStateType()
            state_copy.CopyFrom(state)
            state = state_copy

        # Create context for the authorizer. This is a `ReaderContext`
        # independently of the calling context.
        with self.use_context(
            headers=(
                # Get headers suitable for doing authorization.
                headers.copy_for_token_verification_and_authorization()
            ),
            context_type=rbt.aio.contexts.ReaderContext,
        ) as context:
            context.auth = auth

            # Get the authorizer decision.
            authorization_decision = await self._authorizer.authorize(
                method_name=method_name,
                context=context,
                state=state,
                request=request,
            )

        # Enforce correct authorizer decision type.
        try:
            assert_type(
                authorization_decision,
                [
                    rbt.v1alpha1.errors_pb2.Ok,
                    rbt.v1alpha1.errors_pb2.Unauthenticated,
                    rbt.v1alpha1.errors_pb2.PermissionDenied,
                ]
            )
        except TypeError as e:
            # Retyping.cast the exception to provide more context.
            authorizer_type = f"{type(self._authorizer).__module__}.{type(self._authorizer).__name__}"
            raise TypeError(
                f"Authorizer '{authorizer_type}' "
                f"returned unexpected type '{type(authorization_decision).__name__}' "
                f"for method '{method_name}' on '{headers.state_ref}'"
            ) from e

        # If the decision is not `True`, raise a `SystemAborted` with either a
        # `PermissionDenied` error (in case of `False`) or an `Unauthenticated`
        # error.
        if not isinstance(authorization_decision, rbt.v1alpha1.errors_pb2.Ok):
            if isinstance(authorization_decision, rbt.v1alpha1.errors_pb2.Unauthenticated):
                logger.warning(
                    "Unauthenticated call to "
                    f"'{method_name}' on '{headers.state_ref}'"
                )

            raise rbt.aio.aborted.SystemAborted(
                authorization_decision,
                message=
                f'You are not authorized to call {method_name} on '
                f"'{headers.state_ref}'"
            )

    async def _maybe_verify_token(self, headers) -> typing.Optional[rbt.aio.auth.Auth]:
        """Verify the bearer token and if a token verifier is present.

        Returns the (optional) `rbt.aio.auth.Auth` object
        produced by the token verifier if the token can be verified.
        """
        auth: typing.Optional[rbt.aio.auth.Auth] = None

        if self._token_verifier is not None and headers.bearer_token is not None:
            if headers.bearer_token == __internal_magic_token__:
                return auth

            with self.use_context(
                headers=(
                    # Get headers suitable for doing token verification.
                    headers.copy_for_token_verification_and_authorization()
                ),
                context_type=rbt.aio.contexts.ReaderContext,
            ) as context:
                auth = await self._token_verifier.verify_token(
                    context=context,
                    token=headers.bearer_token,
                )

        return auth


class _APIKeyStub(rbt.aio.stubs.Stub):

    __service_name__ = rbt.aio.types.ServiceName('rbt.cloud.v1alpha1.auth.APIKeyInterface')

    def __init__(
        self,
        *,
        context: rbt.aio.contexts.Context | rbt.aio.external.ExternalContext,
        state_ref: rbt.aio.types.StateRef,
        bearer_token: typing.Optional[str] = None,
    ):
        # ISSUE(https://github.com/reboot-dev/respect/issues/2258) For
        # now we do not explicitly pass on the bearer token unless it
        # is from an `ExternalContext`, as it might have security
        # implications to just pass on the token to any service we are
        # calling.
        if isinstance(context, rbt.aio.external.ExternalContext):
            bearer_token = context.bearer_token

        super().__init__(
            channel_manager=context.channel_manager,
            idempotency_manager=context,
            service_name=self.__service_name__,
            state_ref=state_ref,
            context=context if isinstance(context, rbt.aio.contexts.Context) else None,
            bearer_token=bearer_token,
        )

        channel = self._channel_manager.get_channel_for(
            self.__service_name__, state_ref
        )
        self._stub = rbt.cloud.v1alpha1.auth.auth_pb2_grpc.APIKeyInterfaceStub(channel)


class APIKeyReaderStub(_APIKeyStub):

    def __init__(
        self,
        context: rbt.aio.contexts.ReaderContext | rbt.aio.contexts.WriterContext | rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
        *,
        state_ref: rbt.aio.types.StateRef,
        bearer_token: typing.Optional[str] = None,
    ):
        assert_type(context, [rbt.aio.contexts.ReaderContext, rbt.aio.contexts.WriterContext, rbt.aio.contexts.TransactionContext, rbt.aio.contexts.WorkflowContext, rbt.aio.external.ExternalContext])
        super().__init__(
            context=context,
            state_ref=state_ref,
            bearer_token=bearer_token,
        )

    # APIKey specific methods:


    async def Authenticate(
        self,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest,
        *,
        metadata: typing.Optional[rbt.aio.types.GrpcMetadata] = None,
        bearer_token: typing.Optional[str] = None,
    ) -> rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse:
        async with self._call(
            'Authenticate',
            self._stub.Authenticate,
            request,
            unary=True,
            reader=True,
            response_type=rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse,
            aborted_type=APIKey.AuthenticateAborted,
            metadata=metadata,
            bearer_token=bearer_token,
        ) as call:
            assert isinstance(call, typing.Awaitable)
            return await call



class APIKeyWriterStub(_APIKeyStub):

    def __init__(
        self,
        context: rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
        *,
        state_ref: rbt.aio.types.StateRef,
        bearer_token: typing.Optional[str] = None,
    ):
        assert_type(context, [rbt.aio.contexts.TransactionContext, rbt.aio.contexts.WorkflowContext, rbt.aio.external.ExternalContext])
        super().__init__(
            context=context,
            state_ref=state_ref,
            bearer_token=bearer_token,
        )

    # APIKey specific methods:
    async def Create(
        self,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.CreateRequest,
        idempotency: typing.Optional[rbt.aio.idempotency.Idempotency] = None,
        *,
        metadata: typing.Optional[rbt.aio.types.GrpcMetadata] = None,
        bearer_token: typing.Optional[str] = None,
    ) -> rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse:
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            state_ref=self._headers.state_ref,
            method='Create',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'Create',
                self._stub.Create,
                request,
                unary=True,
                reader=False,
                response_type=rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse,
                aborted_type=APIKey.CreateAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
                bearer_token=bearer_token,
            ) as call:
                assert isinstance(call, typing.Awaitable)
                return await call

    async def Revoke(
        self,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.RevokeRequest,
        idempotency: typing.Optional[rbt.aio.idempotency.Idempotency] = None,
        *,
        metadata: typing.Optional[rbt.aio.types.GrpcMetadata] = None,
        bearer_token: typing.Optional[str] = None,
    ) -> rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse:
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            state_ref=self._headers.state_ref,
            method='Revoke',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'Revoke',
                self._stub.Revoke,
                request,
                unary=True,
                reader=False,
                response_type=rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse,
                aborted_type=APIKey.RevokeAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
                bearer_token=bearer_token,
            ) as call:
                assert isinstance(call, typing.Awaitable)
                return await call

    async def Authenticate(
        self,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest,
        *,
        metadata: typing.Optional[rbt.aio.types.GrpcMetadata] = None,
        bearer_token: typing.Optional[str] = None,
    ) -> rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse:
        async with self._call(
            'Authenticate',
            self._stub.Authenticate,
            request,
            unary=True,
            reader=True,
            response_type=rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse,
            aborted_type=APIKey.AuthenticateAborted,
            metadata=metadata,
            bearer_token=bearer_token,
        ) as call:
            assert isinstance(call, typing.Awaitable)
            return await call


class APIKeyWorkflowStub(_APIKeyStub):

    def __init__(
        self,
        *,
        context: rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
        state_ref: rbt.aio.types.StateRef,
        bearer_token: typing.Optional[str] = None,
    ):
        assert_type(context, [rbt.aio.contexts.TransactionContext, rbt.aio.contexts.WorkflowContext, rbt.aio.external.ExternalContext])
        super().__init__(
            context=context,
            state_ref=state_ref,
            bearer_token=bearer_token,
        )

    # APIKey specific methods:
    async def Create(
        self,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.CreateRequest,
        idempotency: typing.Optional[rbt.aio.idempotency.Idempotency] = None,
        *,
        metadata: typing.Optional[rbt.aio.types.GrpcMetadata] = None,
        bearer_token: typing.Optional[str] = None,
    ) -> rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse:
        idempotency_key: typing.Optional[uuid.UUID]
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            state_ref=self._headers.state_ref,
            method='Create',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'Create',
                self._stub.Create,
                request,
                unary=True,
                reader=False,
                response_type=rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse,
                aborted_type=APIKey.CreateAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
                bearer_token=bearer_token,
            ) as call:
                assert isinstance(call, typing.Awaitable)
                return await call

    async def Revoke(
        self,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.RevokeRequest,
        idempotency: typing.Optional[rbt.aio.idempotency.Idempotency] = None,
        *,
        metadata: typing.Optional[rbt.aio.types.GrpcMetadata] = None,
        bearer_token: typing.Optional[str] = None,
    ) -> rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse:
        idempotency_key: typing.Optional[uuid.UUID]
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            state_ref=self._headers.state_ref,
            method='Revoke',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'Revoke',
                self._stub.Revoke,
                request,
                unary=True,
                reader=False,
                response_type=rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse,
                aborted_type=APIKey.RevokeAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
                bearer_token=bearer_token,
            ) as call:
                assert isinstance(call, typing.Awaitable)
                return await call

    async def Authenticate(
        self,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest,
        *,
        metadata: typing.Optional[rbt.aio.types.GrpcMetadata] = None,
        bearer_token: typing.Optional[str] = None,
    ) -> rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse:
        async with self._call(
            'Authenticate',
            self._stub.Authenticate,
            request,
            unary=True,
            reader=True,
            response_type=rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse,
            aborted_type=APIKey.AuthenticateAborted,
            metadata=metadata,
            bearer_token=bearer_token,
        ) as call:
            assert isinstance(call, typing.Awaitable)
            return await call



class APIKeyTasksStub(_APIKeyStub):

    def __init__(
        self,
        context: rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
        *,
        state_ref: rbt.aio.types.StateRef,
        bearer_token: typing.Optional[str] = None,
    ):
        assert_type(context, [rbt.aio.contexts.TransactionContext, rbt.aio.contexts.WorkflowContext, rbt.aio.external.ExternalContext])
        super().__init__(
            context=context,
            state_ref=state_ref,
            bearer_token=bearer_token,
        )

    # APIKey specific methods:
    async def Revoke(
        self,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.RevokeRequest,
        idempotency: typing.Optional[rbt.aio.idempotency.Idempotency] = None,
        *,
        metadata: typing.Optional[rbt.aio.types.GrpcMetadata] = None,
        bearer_token: typing.Optional[str] = None,
    ) -> tasks_pb2.TaskId:
        idempotency_key: typing.Optional[uuid.UUID]
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            state_ref=self._headers.state_ref,
            method='Revoke',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'Revoke',
                self._stub.Revoke,
                request,
                unary=True,
                reader=False,
                response_type=rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse,
                aborted_type=APIKey.RevokeAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
                bearer_token=bearer_token,
            ) as call:
                assert isinstance(call, typing.Awaitable)
                await call
                for (key, value) in await call.trailing_metadata():  # type: ignore[attr-defined]
                    if key == rbt.aio.headers.TASK_ID_UUID:
                        return tasks_pb2.TaskId(
                            state_type=self._headers.state_type_name,
                            state_ref=self._headers.state_ref.to_str(),
                            task_uuid=uuid.UUID(value).bytes,
                        )
                raise rbt.aio.aborted.SystemAborted(
                    rbt.v1alpha1.errors_pb2.Internal(),
                    message='Trailing metadata missing for task schedule',
                )
    async def Authenticate(
        self,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest,
        idempotency: typing.Optional[rbt.aio.idempotency.Idempotency] = None,
        *,
        metadata: typing.Optional[rbt.aio.types.GrpcMetadata] = None,
        bearer_token: typing.Optional[str] = None,
    ) -> tasks_pb2.TaskId:
        idempotency_key: typing.Optional[uuid.UUID]
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            state_ref=self._headers.state_ref,
            method='Authenticate',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'Authenticate',
                self._stub.Authenticate,
                request,
                unary=True,
                reader=False,
                response_type=rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse,
                aborted_type=APIKey.AuthenticateAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
                bearer_token=bearer_token,
            ) as call:
                assert isinstance(call, typing.Awaitable)
                await call
                for (key, value) in await call.trailing_metadata():  # type: ignore[attr-defined]
                    if key == rbt.aio.headers.TASK_ID_UUID:
                        return tasks_pb2.TaskId(
                            state_type=self._headers.state_type_name,
                            state_ref=self._headers.state_ref.to_str(),
                            task_uuid=uuid.UUID(value).bytes,
                        )
                raise rbt.aio.aborted.SystemAborted(
                    rbt.v1alpha1.errors_pb2.Internal(),
                    message='Trailing metadata missing for task schedule',
                )


class APIKeyServicerTasksStub(_APIKeyStub):

    _context: rbt.aio.contexts.WriterContext

    def __init__(
        self,
        context: rbt.aio.contexts.WriterContext,
        *,
        state_ref: rbt.aio.types.StateRef,
        bearer_token: typing.Optional[str] = None,
    ):
        assert_type(context, [rbt.aio.contexts.WriterContext])
        super().__init__(
            context=context,
            state_ref=state_ref,
            bearer_token=bearer_token,
        )

        self._context = context

    # APIKey specific methods:
    def Create(
        self,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.CreateRequest,
        *,
        schedule: typing.Optional[datetime | timedelta] = None,
    ) -> rbt.aio.tasks.TaskEffect:
        task = rbt.aio.tasks.TaskEffect(
            service_name=self.__service_name__,
            state_ref=self._headers.state_ref,
            method_name='Create',
            request=request,
            schedule=(datetime.now() + schedule) if isinstance(
                schedule, timedelta
            ) else schedule,
        )

        self._context._tasks.append(task)

        return task

    def Revoke(
        self,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.RevokeRequest,
        *,
        schedule: typing.Optional[datetime | timedelta] = None,
    ) -> rbt.aio.tasks.TaskEffect:
        task = rbt.aio.tasks.TaskEffect(
            service_name=self.__service_name__,
            state_ref=self._headers.state_ref,
            method_name='Revoke',
            request=request,
            schedule=(datetime.now() + schedule) if isinstance(
                schedule, timedelta
            ) else schedule,
        )

        self._context._tasks.append(task)

        return task

    def Authenticate(
        self,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest,
        *,
        schedule: typing.Optional[datetime | timedelta] = None,
    ) -> rbt.aio.tasks.TaskEffect:
        task = rbt.aio.tasks.TaskEffect(
            service_name=self.__service_name__,
            state_ref=self._headers.state_ref,
            method_name='Authenticate',
            request=request,
            schedule=(datetime.now() + schedule) if isinstance(
                schedule, timedelta
            ) else schedule,
        )

        self._context._tasks.append(task)

        return task



class APIKeyAuthorizerMixin:
    """Mixin that provides a default implementation of the `can_authorize`
    method for all APIKey authorizers.
    """

    def can_authorize(self, method_name: str) -> bool:
        service_method_names = [
            'rbt.cloud.v1alpha1.auth.APIKey.Create',
            'rbt.cloud.v1alpha1.auth.APIKey.Revoke',
            'rbt.cloud.v1alpha1.auth.APIKey.Authenticate',
        ]
        return method_name in service_method_names


APIKeyStateType: typing.TypeAlias = rbt.cloud.v1alpha1.auth.auth_pb2.APIKey
APIKeyRequestTypes: typing.TypeAlias = \
        rbt.cloud.v1alpha1.auth.auth_pb2.CreateRequest \
        | rbt.cloud.v1alpha1.auth.auth_pb2.RevokeRequest \
        | rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest


class APIKeyAuthorizer(
    APIKeyAuthorizerMixin,
    rbt.aio.auth.authorizers.Authorizer[APIKeyStateType, APIKeyRequestTypes],
):
    StateType: typing.TypeAlias = APIKeyStateType
    RequestTypes: typing.TypeAlias = APIKeyRequestTypes

    async def authorize(
        self,
        *,
        method_name: str,
        context: rbt.aio.contexts.ReaderContext,
        state: typing.Optional[APIKeyStateType] = None,
        request: typing.Optional[APIKeyRequestTypes] = None,
    ) -> rbt.aio.auth.authorizers.Authorizer.Decision:
        if method_name == 'rbt.cloud.v1alpha1.auth.APIKey.Create':
            return await self.Create(
                context,
                typing.cast(rbt.cloud.v1alpha1.auth.auth_pb2.CreateRequest, request),
            )
        elif method_name == 'rbt.cloud.v1alpha1.auth.APIKey.Revoke':
            return await self.Revoke(
                context,
                typing.cast(rbt.cloud.v1alpha1.auth.auth_pb2.APIKey, state),
                typing.cast(rbt.cloud.v1alpha1.auth.auth_pb2.RevokeRequest, request),
            )
        elif method_name == 'rbt.cloud.v1alpha1.auth.APIKey.Authenticate':
            return await self.Authenticate(
                context,
                typing.cast(rbt.cloud.v1alpha1.auth.auth_pb2.APIKey, state),
                typing.cast(rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest, request),
            )
        else:
            return rbt.v1alpha1.errors_pb2.PermissionDenied()

    async def Create(
        self,
        context: rbt.aio.contexts.ReaderContext,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.CreateRequest,
    ) -> rbt.aio.auth.authorizers.Authorizer.Decision:
        return rbt.v1alpha1.errors_pb2.PermissionDenied()

    async def Revoke(
        self,
        context: rbt.aio.contexts.ReaderContext,
        state: rbt.cloud.v1alpha1.auth.auth_pb2.APIKey,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.RevokeRequest,
    ) -> rbt.aio.auth.authorizers.Authorizer.Decision:
        return rbt.v1alpha1.errors_pb2.PermissionDenied()

    async def Authenticate(
        self,
        context: rbt.aio.contexts.ReaderContext,
        state: rbt.cloud.v1alpha1.auth.auth_pb2.APIKey,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest,
    ) -> rbt.aio.auth.authorizers.Authorizer.Decision:
        return rbt.v1alpha1.errors_pb2.PermissionDenied()


    # Off-the shelf authorizers for APIKey:

    class AllowAllIfAuthenticated(
        APIKeyAuthorizerMixin,
        rbt.aio.auth.authorizers.Authorizer,
    ):
        """An authorizer for APIKey that allows all requests if the
        caller is authenticated.
        """
        async def authorize(
            self,
            *,
            method_name: str,
            context: rbt.aio.contexts.ReaderContext,
            state: typing.Optional[APIKeyStateType] = None,
            request: typing.Optional[APIKeyRequestTypes] = None,
        ) -> rbt.aio.auth.authorizers.Authorizer.Decision:
            if context.auth is None:
                return rbt.v1alpha1.errors_pb2.Unauthenticated()

            return rbt.v1alpha1.errors_pb2.Ok()


class APIKeyServicer(rbt.aio.servicers.Servicer):
    Authorizer: typing.TypeAlias = APIKeyAuthorizer

    __service_name__ = rbt.aio.types.ServiceName('rbt.cloud.v1alpha1.auth.APIKeyInterface')
    __state_type_name__ = rbt.aio.types.StateTypeName('rbt.cloud.v1alpha1.auth.APIKey')
    __state_type__ = rbt.cloud.v1alpha1.auth.auth_pb2.APIKey
    __file_descriptor__ = rbt.cloud.v1alpha1.auth.auth_pb2.DESCRIPTOR

    def __init__(self):
        super().__init__()
        # NOTE: need to hold on to the middleware so we can do inline
        # writes (see 'self.write(...)').
        #
        # Because '_middleware' is not really private this does mean
        # users may do possibly dangerous things, but this is no more
        # likely given they could have already overridden
        # 'create_middleware()'.
        self._middleware: typing.Optional[APIKeyServicerMiddleware] = None

    def create_middleware(
        self,
        *,
        application_id: rbt.aio.types.ApplicationId,
        consensus_id: rbt.aio.types.ConsensusId,
        state_manager: rbt.aio.state_managers.StateManager,
        placement_client: rbt.aio.placement.PlacementClient,
        channel_manager: rbt.aio.internals.channel_manager._ChannelManager,
        tasks_cache: rbt.aio.internals.tasks_cache.TasksCache,
        token_verifier: typing.Optional[rbt.aio.auth.token_verifiers.TokenVerifier],
        effect_validation: rbt.aio.contexts.EffectValidation,
        app_internal_api_key_secret: str,
        ready: asyncio.Event,
    ) -> APIKeyServicerMiddleware:
        self._middleware = APIKeyServicerMiddleware(
            servicer=self,
            application_id=application_id,
            consensus_id=consensus_id,
            state_manager=state_manager,
            placement_client=placement_client,
            channel_manager=channel_manager,
            tasks_cache=tasks_cache,
            token_verifier=token_verifier,
            effect_validation=effect_validation,
            app_internal_api_key_secret=app_internal_api_key_secret,
            ready=ready,
        )
        return self._middleware

    def authorizer(self) -> typing.Optional[rbt.aio.auth.authorizers.Authorizer]:
        return None

    def token_verifier(self) -> typing.Optional[rbt.aio.auth.token_verifiers.TokenVerifier]:
        return None

    def lookup(
        self,
        *,
        bearer_token: typing.Optional[str] = None,
    ) -> APIKey.WeakReference[APIKey.WeakReference._WriterSchedule]:
        # TODO(benh): it's still possible that the `bearer_token` from
        # the context will not be sufficient for making a
        # call/schedule due to a user defined authorizer; consider
        # creating a "per middleware" bearer token that always works
        # for calls from `self.lookup()`.
        context = rbt.aio.contexts.Context.get()

        if context is None:
            raise RuntimeError(
                'Missing asyncio context variable `context`; '
                'are you using this class without Resemble?'
            )

        return APIKey.WeakReference(
            state_id=context._state_ref.id,
            schedule_type=APIKey.WeakReference._WriterSchedule,
            bearer_token=bearer_token or context.bearer_token,
        )

    class Effects(rbt.aio.state_managers.Effects):
        def __init__(
            self,
            *,
            state: rbt.cloud.v1alpha1.auth.auth_pb2.APIKey,
            response: typing.Optional[google.protobuf.message.Message] = None,
            tasks: typing.Optional[list[rbt.aio.tasks.TaskEffect]] = None,
            _colocated_upserts: typing.Optional[list[tuple[str, typing.Optional[bytes]]]] = None,
        ):
            assert_type(state, [rbt.cloud.v1alpha1.auth.auth_pb2.APIKey])

            super().__init__(state=state, response=response, tasks=tasks, _colocated_upserts=_colocated_upserts)

    class CreateEffects(Effects):
        def __init__(
            self,
            *,
            state: rbt.cloud.v1alpha1.auth.auth_pb2.APIKey,
            response: rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse,
            tasks: typing.Optional[list[rbt.aio.tasks.TaskEffect]] = None,
            _colocated_upserts: typing.Optional[list[tuple[str, typing.Optional[bytes]]]] = None,
        ):
            assert_type(state, [rbt.cloud.v1alpha1.auth.auth_pb2.APIKey])
            assert_type(response, [rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse])

            super().__init__(state=state, response=response, tasks=tasks, _colocated_upserts=_colocated_upserts)


    class RevokeEffects(Effects):
        def __init__(
            self,
            *,
            state: rbt.cloud.v1alpha1.auth.auth_pb2.APIKey,
            response: rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse,
            tasks: typing.Optional[list[rbt.aio.tasks.TaskEffect]] = None,
            _colocated_upserts: typing.Optional[list[tuple[str, typing.Optional[bytes]]]] = None,
        ):
            assert_type(state, [rbt.cloud.v1alpha1.auth.auth_pb2.APIKey])
            assert_type(response, [rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse])

            super().__init__(state=state, response=response, tasks=tasks, _colocated_upserts=_colocated_upserts)




    class InlineWriterCallable(typing.Protocol):
        async def __call__(
            self,
            context: rbt.aio.contexts.WriterContext,
            state: rbt.cloud.v1alpha1.auth.auth_pb2.APIKey
        ) -> APIKeyServicer.Effects:
            ...

    async def read(
        self, context: rbt.aio.contexts.WorkflowContext
    ) -> rbt.cloud.v1alpha1.auth.auth_pb2.APIKey:
        """Read the current state within a transaction."""

        assert_type(context, [rbt.aio.contexts.WorkflowContext])

        if self._middleware is None:
            raise RuntimeError(
                'Resemble middleware was not created; '
                'are you using this class without Resemble?'
            )

        return await self._middleware._state_manager.read(
            context, self.__state_type__
        )

    async def write(
        self,
        context: rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext,
        writer: InlineWriterCallable,
        __options__: typing.Optional[rbt.aio.call.Options] = None,
    ):
        """Perform an "inline write" within a transaction or workflow."""
        assert_type(context, [rbt.aio.contexts.TransactionContext, rbt.aio.contexts.WorkflowContext])

        if self._middleware is None:
            raise RuntimeError(
                'Resemble middleware was not created; '
                'are you using this class without Resemble?'
            )

        idempotency: typing.Optional[rbt.aio.idempotency.Idempotency] = None
        metadata: typing.Optional[rbt.aio.types.GrpcMetadata] = None

        if __options__ is not None:
            idempotency = __options__.idempotency
            if __options__.metadata is not None:
                metadata = __options__.metadata

        if metadata is None:
            metadata = ()

        headers = rbt.aio.headers.Headers(
            application_id=context.application_id,
            service_name=self.__service_name__,
            state_ref=context._state_ref,
            transaction_ids=context.transaction_ids,
            transaction_coordinator_service=context
            .transaction_coordinator_service,
            transaction_coordinator_state_ref=context
            .transaction_coordinator_state_ref
        )

        metadata += headers.to_grpc_metadata()

        idempotency_key: typing.Optional[uuid.UUID]
        with context.idempotently(
            service=self.__service_name__,
            state_ref=context._state_ref,
            method=None,
            request=None,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:

            if any(t[0] == rbt.aio.headers.IDEMPOTENCY_KEY_HEADER for t in metadata):
                raise ValueError(
                    f"Do not set '{rbt.aio.headers.IDEMPOTENCY_KEY_HEADER}' metadata yourself"
                )

            if idempotency_key is not None:
                metadata += (
                    (rbt.aio.headers.IDEMPOTENCY_KEY_HEADER, str(idempotency_key)),
                )

            with self._middleware.use_context(
                headers=rbt.aio.headers.Headers.from_grpc_metadata(metadata),
                context_type=rbt.aio.contexts.WriterContext,
            ) as writer_context:
                # Check if we already have performed this mutation!
                #
                # We do this _before_ calling 'transactionally()' because
                # if this call is for a transaction method _and_ we've
                # already performed the transaction then we don't want to
                # become a transaction participant (again) we just want to
                # return the transaction's response.
                idempotent_mutation = (
                    self._middleware._state_manager.check_for_idempotent_mutation(
                        writer_context
                    )
                )

                if idempotent_mutation is not None:
                    # NOTE: invariant for now is that inline writers can't
                    # have responses which we also check for below.
                    assert len(idempotent_mutation.response) == 0
                    return

                async with self._middleware._state_manager.transactionally(
                    writer_context, self._middleware.tasks_dispatcher
                ) as transaction:
                    async with self._middleware._state_manager.writer(
                        writer_context,
                        self.__state_type__,
                        self._middleware.tasks_dispatcher,
                        transaction=transaction,
                    ) as (state, state_manager_writer):
                        await writer(context=writer_context, state=state)

                        effects = APIKeyServicer.Effects(
                            state=state,
                            tasks=context._tasks,
                        )

                        # All idempotent mutations require a reponse, so
                        # for now we just use `Empty`.
                        #
                        # TODO(benh): figure out if we want to support
                        # returning a response because we'd need to
                        # pickle it maybe using `memoize` to support
                        # an idempotent mutation.
                        assert effects.response is None
                        effects.response = empty_pb2.Empty()

                        await state_manager_writer.complete(effects)

    # APIKey specific methods:
    class _Idempotently:

        def __init__(
            self,
            *,
            servicer: APIKeyServicer,
            idempotency: rbt.aio.idempotency.Idempotency,
        ):
            self._servicer = servicer
            self._idempotency = idempotency

        async def write(
            self,
            context: rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext,
            writer: APIKeyServicer.InlineWriterCallable,
            __options__: rbt.aio.call.Options = rbt.aio.call.Options(),
        ):
            if __options__.idempotency is not None:
                raise RuntimeError(
                    'Found redundant idempotency in `Options`'
                )

            __options__ = dataclasses.replace(
                __options__,
                idempotency_key=self._idempotency.key,
                idempotency_alias=self._idempotency.alias,
                generate_idempotency=self._idempotency.generate,
            )

            return await self._servicer.write(context, writer, __options__)

    def idempotently(
        self,
        alias: typing.Optional[str] = None,
        *,
        key: typing.Optional[uuid.UUID] = None,
    ) -> APIKeyServicer._Idempotently:
        return APIKeyServicer._Idempotently(
            servicer=self,
            idempotency=rbt.aio.idempotency.Idempotency(key=key, alias=alias),
        )

    # APIKey specific methods:
    @abstractmethod
    async def Create(
        self,
        context: rbt.aio.contexts.WriterContext,
        state: rbt.cloud.v1alpha1.auth.auth_pb2.APIKey,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.CreateRequest,
    ) -> rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse:
        raise NotImplementedError

    @abstractmethod
    async def Revoke(
        self,
        context: rbt.aio.contexts.WriterContext,
        state: rbt.cloud.v1alpha1.auth.auth_pb2.APIKey,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.RevokeRequest,
    ) -> rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse:
        raise NotImplementedError

    @abstractmethod
    async def Authenticate(
        self,
        context: rbt.aio.contexts.ReaderContext,
        state: rbt.cloud.v1alpha1.auth.auth_pb2.APIKey,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest,
    ) -> rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse:
        raise NotImplementedError



APIKey_ScheduleTypeVar = typing.TypeVar('APIKey_ScheduleTypeVar', 'APIKey.WeakReference._Schedule', 'APIKey.WeakReference._WriterSchedule')
APIKey_IdempotentlyScheduleTypeVar = typing.TypeVar('APIKey_IdempotentlyScheduleTypeVar', 'APIKey.WeakReference._Schedule', 'APIKey.WeakReference._WriterSchedule')


class APIKey:

    Interface: typing.TypeAlias = APIKeyServicer

    Effects: typing.TypeAlias = APIKeyServicer.Effects

    Authorizer: typing.TypeAlias = APIKeyAuthorizer

    State: typing.TypeAlias = rbt.cloud.v1alpha1.auth.auth_pb2.APIKey

    __state_type_name__ = rbt.aio.types.StateTypeName("rbt.cloud.v1alpha1.auth.APIKey")

    class CreateFuture:
        """Represents a future corresponding to a task running for the
        state. Note that this is not a coroutine because we are trying
        to convey the semantics that the task is already running (or
        will soon be) and thus we are just giving you a (distributed)
        future to it.
        """

        def __init__(
            self,
            context: rbt.aio.contexts.Context | rbt.aio.external.ExternalContext,
            *,
            task_id: tasks_pb2.TaskId,
        ) -> None:
            self._channel_manager = context.channel_manager
            self._task_id = task_id

        @property
        def task_id(self) -> tasks_pb2.TaskId:
            return self._task_id

        def __await__(self) -> typing.Generator[None, None, rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse]:
            """Awaits for task to finish and returns it's response."""

            async def wait_for_task() -> rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse:
                channel = self._channel_manager.get_channel_from_state_type_name(
                    self._task_id.state_type,
                    rbt.aio.types.StateRef(self._task_id.state_ref),
                )

                stub = tasks_pb2_grpc.TasksStub(channel)

                try:
                    wait_for_task_response = await stub.Wait(
                        tasks_pb2.WaitRequest(task_id=self._task_id),
                        metadata=rbt.aio.headers.Headers(
                            service_name=state_type_to_service(self._task_id.state_type),
                            state_ref=rbt.aio.types.StateRef(self._task_id.state_ref),
                            # TODO(benh): will eventually be necessary to
                            # disambiguate between applications.
                            application_id=None,
                        ).to_grpc_metadata(),
                    )
                except grpc.aio.AioRpcError as error:
                    if error.code() == grpc.StatusCode.NOT_FOUND:
                        raise rbt.aio.aborted.SystemAborted(
                            rbt.v1alpha1.errors_pb2.UnknownTask()
                        ) from None

                    raise rbt.aio.aborted.SystemAborted.from_grpc_aio_rpc_error(
                        error
                    ) from None
                else:
                    response_or_error: typing.Optional[google.protobuf.message.Message] = None
                    is_error = False

                    if wait_for_task_response.response_or_error.WhichOneof("response_or_error") == "response":
                        response_or_error = wait_for_task_response.response_or_error.response
                    else:
                        is_error = True
                        response_or_error = wait_for_task_response.response_or_error.error

                    response = rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse()
                    if (
                        not is_error and response_or_error.TypeName() != response.DESCRIPTOR.full_name
                    ):
                        raise rbt.aio.aborted.SystemAborted(
                            rbt.v1alpha1.errors_pb2.InvalidArgument(),
                            message=
                            f"task with UUID {str(uuid.UUID(bytes=self._task_id.task_uuid))} "
                            f"has a response of type '{response_or_error.TypeName()}' "
                            "but expecting type 'rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse'; "
                            "are you waiting on a task of the correct method?",
                        ) from None

                    if is_error:
                        # Currently only cancelled error is supported.
                        raise asyncio.CancelledError("Task was cancelled by a TasksServicer")
                    else:
                        response_or_error.Unpack(response)
                        return response

            return wait_for_task().__await__()

    CreateEffects: typing.TypeAlias = Interface.CreateEffects

    class CreateAborted(rbt.aio.aborted.Aborted):

        MethodError = typing.Union[
            rbt.cloud.v1alpha1.auth.auth_pb2.InvalidInputError        ]

        Error = typing.Union[
            MethodError,
            rbt.aio.aborted.GrpcError,
            rbt.aio.aborted.ResembleError,
        ]

        METHOD_ERROR_TYPES: list[type[Error]] = [
            rbt.cloud.v1alpha1.auth.auth_pb2.InvalidInputError        ]

        ERROR_TYPES: list[type[Error]] = (
            METHOD_ERROR_TYPES +
            rbt.aio.aborted.GRPC_ERROR_TYPES +
            rbt.aio.aborted.RESEMBLE_ERROR_TYPES
        )

        _error: Error
        _code: grpc.StatusCode
        _message: typing.Optional[str]

        def __init__(
            self,
            error: MethodError |  rbt.aio.aborted.GrpcError,
            *,
            message: typing.Optional[str] = None,
            # Do not set this value when constructing in order to
            # raise. This is only used internally when constructing
            # from aborted calls.
            error_types: list[type[Error]] = (
                METHOD_ERROR_TYPES + rbt.aio.aborted.GRPC_ERROR_TYPES
            ),
        ):
            super().__init__()

            assert_type(error, error_types)

            self._error = error

            code = self.grpc_status_code_from_error(self._error)

            if code is None:
                # Must be a Resemble specific or declared method error.
                code = grpc.StatusCode.ABORTED

            self._code = code

            self._message = message

        @property
        def error(self) -> Error:
            return self._error

        @property
        def code(self) -> grpc.StatusCode:
            return self._code

        @property
        def message(self) -> typing.Optional[str]:
            return self._message

        @classmethod
        def from_status(cls, status: status_pb2.Status):
            error = cls.error_from_google_rpc_status_details(
                status,
                cls.ERROR_TYPES,
            )

            message = status.message if len(status.message) > 0 else None

            if error is not None:
                return cls(error, message=message, error_types=cls.ERROR_TYPES)

            error = cls.error_from_google_rpc_status_code(status)

            assert error is not None

            # TODO(benh): also consider getting the type names from
            # `status.details` and including that in `message` to make
            # debugging easier.

            return cls(error, message=message)

        @classmethod
        def from_grpc_aio_rpc_error(cls, aio_rpc_error: grpc.aio.AioRpcError):
            return cls(
                cls.error_from_grpc_aio_rpc_error(aio_rpc_error),
                message=aio_rpc_error.details(),
            )

        @classmethod
        def is_declared_error(cls, message: google.protobuf.message.Message) -> bool:
            if message.DESCRIPTOR.full_name == 'rbt.cloud.v1alpha1.auth.InvalidInputError':
                return True
            return False

    class RevokeFuture:
        """Represents a future corresponding to a task running for the
        state. Note that this is not a coroutine because we are trying
        to convey the semantics that the task is already running (or
        will soon be) and thus we are just giving you a (distributed)
        future to it.
        """

        def __init__(
            self,
            context: rbt.aio.contexts.Context | rbt.aio.external.ExternalContext,
            *,
            task_id: tasks_pb2.TaskId,
        ) -> None:
            self._channel_manager = context.channel_manager
            self._task_id = task_id

        @property
        def task_id(self) -> tasks_pb2.TaskId:
            return self._task_id

        def __await__(self) -> typing.Generator[None, None, rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse]:
            """Awaits for task to finish and returns it's response."""

            async def wait_for_task() -> rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse:
                channel = self._channel_manager.get_channel_from_state_type_name(
                    self._task_id.state_type,
                    rbt.aio.types.StateRef(self._task_id.state_ref),
                )

                stub = tasks_pb2_grpc.TasksStub(channel)

                try:
                    wait_for_task_response = await stub.Wait(
                        tasks_pb2.WaitRequest(task_id=self._task_id),
                        metadata=rbt.aio.headers.Headers(
                            service_name=state_type_to_service(self._task_id.state_type),
                            state_ref=rbt.aio.types.StateRef(self._task_id.state_ref),
                            # TODO(benh): will eventually be necessary to
                            # disambiguate between applications.
                            application_id=None,
                        ).to_grpc_metadata(),
                    )
                except grpc.aio.AioRpcError as error:
                    if error.code() == grpc.StatusCode.NOT_FOUND:
                        raise rbt.aio.aborted.SystemAborted(
                            rbt.v1alpha1.errors_pb2.UnknownTask()
                        ) from None

                    raise rbt.aio.aborted.SystemAborted.from_grpc_aio_rpc_error(
                        error
                    ) from None
                else:
                    response_or_error: typing.Optional[google.protobuf.message.Message] = None
                    is_error = False

                    if wait_for_task_response.response_or_error.WhichOneof("response_or_error") == "response":
                        response_or_error = wait_for_task_response.response_or_error.response
                    else:
                        is_error = True
                        response_or_error = wait_for_task_response.response_or_error.error

                    response = rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse()
                    if (
                        not is_error and response_or_error.TypeName() != response.DESCRIPTOR.full_name
                    ):
                        raise rbt.aio.aborted.SystemAborted(
                            rbt.v1alpha1.errors_pb2.InvalidArgument(),
                            message=
                            f"task with UUID {str(uuid.UUID(bytes=self._task_id.task_uuid))} "
                            f"has a response of type '{response_or_error.TypeName()}' "
                            "but expecting type 'rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse'; "
                            "are you waiting on a task of the correct method?",
                        ) from None

                    if is_error:
                        # Currently only cancelled error is supported.
                        raise asyncio.CancelledError("Task was cancelled by a TasksServicer")
                    else:
                        response_or_error.Unpack(response)
                        return response

            return wait_for_task().__await__()

    RevokeEffects: typing.TypeAlias = Interface.RevokeEffects

    class RevokeAborted(rbt.aio.aborted.Aborted):


        Error = typing.Union[
            rbt.aio.aborted.GrpcError,
            rbt.aio.aborted.ResembleError,
        ]

        METHOD_ERROR_TYPES: list[type[Error]] = [
        ]

        ERROR_TYPES: list[type[Error]] = (
            METHOD_ERROR_TYPES +
            rbt.aio.aborted.GRPC_ERROR_TYPES +
            rbt.aio.aborted.RESEMBLE_ERROR_TYPES
        )

        _error: Error
        _code: grpc.StatusCode
        _message: typing.Optional[str]

        def __init__(
            self,
            error:  rbt.aio.aborted.GrpcError,
            *,
            message: typing.Optional[str] = None,
            # Do not set this value when constructing in order to
            # raise. This is only used internally when constructing
            # from aborted calls.
            error_types: list[type[Error]] = (
                METHOD_ERROR_TYPES + rbt.aio.aborted.GRPC_ERROR_TYPES
            ),
        ):
            super().__init__()

            assert_type(error, error_types)

            self._error = error

            code = self.grpc_status_code_from_error(self._error)

            if code is None:
                # Must be a Resemble specific or declared method error.
                code = grpc.StatusCode.ABORTED

            self._code = code

            self._message = message

        @property
        def error(self) -> Error:
            return self._error

        @property
        def code(self) -> grpc.StatusCode:
            return self._code

        @property
        def message(self) -> typing.Optional[str]:
            return self._message

        @classmethod
        def from_status(cls, status: status_pb2.Status):
            error = cls.error_from_google_rpc_status_details(
                status,
                cls.ERROR_TYPES,
            )

            message = status.message if len(status.message) > 0 else None

            if error is not None:
                return cls(error, message=message, error_types=cls.ERROR_TYPES)

            error = cls.error_from_google_rpc_status_code(status)

            assert error is not None

            # TODO(benh): also consider getting the type names from
            # `status.details` and including that in `message` to make
            # debugging easier.

            return cls(error, message=message)

        @classmethod
        def from_grpc_aio_rpc_error(cls, aio_rpc_error: grpc.aio.AioRpcError):
            return cls(
                cls.error_from_grpc_aio_rpc_error(aio_rpc_error),
                message=aio_rpc_error.details(),
            )

        @classmethod
        def is_declared_error(cls, message: google.protobuf.message.Message) -> bool:
            return False

    class AuthenticateFuture:
        """Represents a future corresponding to a task running for the
        state. Note that this is not a coroutine because we are trying
        to convey the semantics that the task is already running (or
        will soon be) and thus we are just giving you a (distributed)
        future to it.
        """

        def __init__(
            self,
            context: rbt.aio.contexts.Context | rbt.aio.external.ExternalContext,
            *,
            task_id: tasks_pb2.TaskId,
        ) -> None:
            self._channel_manager = context.channel_manager
            self._task_id = task_id

        @property
        def task_id(self) -> tasks_pb2.TaskId:
            return self._task_id

        def __await__(self) -> typing.Generator[None, None, rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse]:
            """Awaits for task to finish and returns it's response."""

            async def wait_for_task() -> rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse:
                channel = self._channel_manager.get_channel_from_state_type_name(
                    self._task_id.state_type,
                    rbt.aio.types.StateRef(self._task_id.state_ref),
                )

                stub = tasks_pb2_grpc.TasksStub(channel)

                try:
                    wait_for_task_response = await stub.Wait(
                        tasks_pb2.WaitRequest(task_id=self._task_id),
                        metadata=rbt.aio.headers.Headers(
                            service_name=state_type_to_service(self._task_id.state_type),
                            state_ref=rbt.aio.types.StateRef(self._task_id.state_ref),
                            # TODO(benh): will eventually be necessary to
                            # disambiguate between applications.
                            application_id=None,
                        ).to_grpc_metadata(),
                    )
                except grpc.aio.AioRpcError as error:
                    if error.code() == grpc.StatusCode.NOT_FOUND:
                        raise rbt.aio.aborted.SystemAborted(
                            rbt.v1alpha1.errors_pb2.UnknownTask()
                        ) from None

                    raise rbt.aio.aborted.SystemAborted.from_grpc_aio_rpc_error(
                        error
                    ) from None
                else:
                    response_or_error: typing.Optional[google.protobuf.message.Message] = None
                    is_error = False

                    if wait_for_task_response.response_or_error.WhichOneof("response_or_error") == "response":
                        response_or_error = wait_for_task_response.response_or_error.response
                    else:
                        is_error = True
                        response_or_error = wait_for_task_response.response_or_error.error

                    response = rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse()
                    if (
                        not is_error and response_or_error.TypeName() != response.DESCRIPTOR.full_name
                    ):
                        raise rbt.aio.aborted.SystemAborted(
                            rbt.v1alpha1.errors_pb2.InvalidArgument(),
                            message=
                            f"task with UUID {str(uuid.UUID(bytes=self._task_id.task_uuid))} "
                            f"has a response of type '{response_or_error.TypeName()}' "
                            "but expecting type 'rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse'; "
                            "are you waiting on a task of the correct method?",
                        ) from None

                    if is_error:
                        # Currently only cancelled error is supported.
                        raise asyncio.CancelledError("Task was cancelled by a TasksServicer")
                    else:
                        response_or_error.Unpack(response)
                        return response

            return wait_for_task().__await__()


    class AuthenticateAborted(rbt.aio.aborted.Aborted):


        Error = typing.Union[
            rbt.aio.aborted.GrpcError,
            rbt.aio.aborted.ResembleError,
        ]

        METHOD_ERROR_TYPES: list[type[Error]] = [
        ]

        ERROR_TYPES: list[type[Error]] = (
            METHOD_ERROR_TYPES +
            rbt.aio.aborted.GRPC_ERROR_TYPES +
            rbt.aio.aborted.RESEMBLE_ERROR_TYPES
        )

        _error: Error
        _code: grpc.StatusCode
        _message: typing.Optional[str]

        def __init__(
            self,
            error:  rbt.aio.aborted.GrpcError,
            *,
            message: typing.Optional[str] = None,
            # Do not set this value when constructing in order to
            # raise. This is only used internally when constructing
            # from aborted calls.
            error_types: list[type[Error]] = (
                METHOD_ERROR_TYPES + rbt.aio.aborted.GRPC_ERROR_TYPES
            ),
        ):
            super().__init__()

            assert_type(error, error_types)

            self._error = error

            code = self.grpc_status_code_from_error(self._error)

            if code is None:
                # Must be a Resemble specific or declared method error.
                code = grpc.StatusCode.ABORTED

            self._code = code

            self._message = message

        @property
        def error(self) -> Error:
            return self._error

        @property
        def code(self) -> grpc.StatusCode:
            return self._code

        @property
        def message(self) -> typing.Optional[str]:
            return self._message

        @classmethod
        def from_status(cls, status: status_pb2.Status):
            error = cls.error_from_google_rpc_status_details(
                status,
                cls.ERROR_TYPES,
            )

            message = status.message if len(status.message) > 0 else None

            if error is not None:
                return cls(error, message=message, error_types=cls.ERROR_TYPES)

            error = cls.error_from_google_rpc_status_code(status)

            assert error is not None

            # TODO(benh): also consider getting the type names from
            # `status.details` and including that in `message` to make
            # debugging easier.

            return cls(error, message=message)

        @classmethod
        def from_grpc_aio_rpc_error(cls, aio_rpc_error: grpc.aio.AioRpcError):
            return cls(
                cls.error_from_grpc_aio_rpc_error(aio_rpc_error),
                message=aio_rpc_error.details(),
            )

        @classmethod
        def is_declared_error(cls, message: google.protobuf.message.Message) -> bool:
            return False


    class WeakReference(typing.Generic[APIKey_ScheduleTypeVar]):

        _schedule_type: type[APIKey_ScheduleTypeVar]

        def __init__(
            self,
            state_id: rbt.aio.types.StateId,
            *,
            schedule_type: type[APIKey_ScheduleTypeVar],
            bearer_token: typing.Optional[str] = None,
        ):
            self._state_ref = rbt.aio.types.StateRef.from_id(
              APIKey.__state_type_name__,
              state_id,
            )
            self._schedule_type = schedule_type
            self._idempotency_manager: typing.Optional[rbt.aio.idempotency.IdempotencyManager] = None
            self._reader_stub: typing.Optional[APIKeyReaderStub] = None
            self._writer_stub: typing.Optional[APIKeyWriterStub] = None
            self._workflow_stub: typing.Optional[APIKeyWorkflowStub] = None
            self._tasks_stub: typing.Optional[APIKeyTasksStub] = None
            self._bearer_token = bearer_token

        @property
        def state_id(self) -> rbt.aio.types.StateId:
            return self._state_ref.id

        def _reader(
            self,
            context: rbt.aio.contexts.ReaderContext | rbt.aio.contexts.WriterContext | rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
        ) -> APIKeyReaderStub:
            if self._reader_stub is None:
                self._reader_stub = APIKeyReaderStub(
                    context=context,
                    state_ref=self._state_ref,
                    bearer_token=self._bearer_token,
                )
            assert self._reader_stub is not None
            if self._idempotency_manager is None:
                self._idempotency_manager = context
            elif self._idempotency_manager != context:
                raise rbt.aio.call.MixedContextsError(
                    "This `WeakReference` for `APIKey` with ID "
                    f"'{self.state_id}' has previously been used by a "
                    "different `Context`. That is not allowed. "
                    "Instead create a new `WeakReference` for every `Context` by calling "
                    f"`APIKey.lookup('{self.state_id}')`."
                )
            return self._reader_stub

        def _writer(
            self,
            context: rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
        ) -> APIKeyWriterStub:
            if self._writer_stub is None:
                self._writer_stub = APIKeyWriterStub(
                    context=context,
                    state_ref=self._state_ref,
                    bearer_token=self._bearer_token,
                )
            assert self._writer_stub is not None
            if self._idempotency_manager is None:
                self._idempotency_manager = context
            elif self._idempotency_manager != context:
                raise rbt.aio.call.MixedContextsError(
                    "This `WeakReference` for `APIKey` with ID "
                    f"'{self.state_id}' has previously been used by a "
                    "different `Context`. That is not allowed. "
                    "Instead create a new `WeakReference` for every `Context` by calling "
                    f"`APIKey.lookup('{self.state_id}')`."
                )
            return self._writer_stub

        def _workflow(
            self,
            context: rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
        ) -> APIKeyWorkflowStub:
            if self._workflow_stub is None:
                self._workflow_stub = APIKeyWorkflowStub(
                    context=context,
                    state_ref=self._state_ref,
                    bearer_token=self._bearer_token,
                )
            assert self._workflow_stub is not None
            if self._idempotency_manager is None:
                self._idempotency_manager = context
            elif self._idempotency_manager != context:
                raise rbt.aio.call.MixedContextsError(
                    "This `WeakReference` for `APIKey` with ID "
                    f"'{self.state_id}' has previously been used by a "
                    "different `Context`. That is not allowed. "
                    "Instead create a new `WeakReference` for every `Context` by calling "
                    f"`APIKey.lookup('{self.state_id}')`."
                )
            return self._workflow_stub

        def _tasks(
            self,
            context: rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
        ) -> APIKeyTasksStub:
            if self._tasks_stub is None:
                self._tasks_stub = APIKeyTasksStub(
                    context=context,
                    state_ref=self._state_ref,
                    bearer_token=self._bearer_token,
                )
            assert self._tasks_stub is not None
            if self._idempotency_manager is None:
                self._idempotency_manager = context
            elif self._idempotency_manager != context:
                raise rbt.aio.call.MixedContextsError(
                    "This `WeakReference` for `APIKey` with ID "
                    f"'{self.state_id}' has previously been used by a "
                    "different `Context`. That is not allowed. "
                    "Instead create a new `WeakReference` for every `Context` by calling "
                    f"`APIKey.lookup('{self.state_id}')`."
                )
            return self._tasks_stub

        class _Reactively:

            def __init__(
                self,
                *,
                service_name: rbt.aio.types.ServiceName,
                state_ref: rbt.aio.types.StateRef,
                bearer_token: typing.Optional[str] = None,
            ):
                self._service_name = service_name
                self._state_ref = state_ref
                self._bearer_token = bearer_token

            async def Authenticate(
                self,
                context: rbt.aio.external.ExternalContext,
                __options__: typing.Optional[rbt.aio.call.Options] = None,
                *,
                secret: typing.Optional[str] = None,
            ) -> typing.AsyncIterator[rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse]:

                assert_type(context, [rbt.aio.external.ExternalContext])

                # TODO: mypy-protobuf declares that
                # `google.protobuf.message.Message` constructor arguments are
                # always non-None, when in reality they are optional.
                request = rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest(
                    secret=secret,  # type: ignore[arg-type]
                )

                metadata: typing.Optional[rbt.aio.types.GrpcMetadata] = None
                bearer_token: typing.Optional[str] = None

                if __options__ is not None:
                    if __options__.metadata is not None:
                        metadata = __options__.metadata
                    if __options__.bearer_token is not None:
                        bearer_token = __options__.bearer_token

                if bearer_token is None:
                    bearer_token = self._bearer_token

                if metadata is None:
                    metadata = ()

                headers = rbt.aio.headers.Headers(
                    bearer_token=bearer_token,
                    service_name=self._service_name,
                    state_ref=self._state_ref,
                    # TODO(benh): will eventually be necessary to
                    # disambiguate between applications.
                    application_id=None,
                )

                metadata += headers.to_grpc_metadata()

                query_backoff = backoff.Backoff()
                while True:
                    try:
                        async with context.channel_manager.get_channel_from_service_name(
                            self._service_name, self._state_ref
                        ) as channel:

                            call = rbt.v1alpha1.react_pb2_grpc.ReactStub(
                                channel
                            ).Query(
                                rbt.v1alpha1.react_pb2.QueryRequest(
                                    method='Authenticate',
                                    request=request.SerializeToString(),
                                ),
                                metadata=metadata,
                            )

                            async for query_response in call:
                                # Clear the backoff so we don't wait
                                # as long the next time we get
                                # disconnected.
                                query_backoff.clear()

                                # The backend may have sent us this query
                                # response only to let us know that a new
                                # idempotency key has been recorded; there may
                                # not be a new response. Python callers don't
                                # (currently) care about such an event, so we
                                # simply ignore it.
                                if not query_response.HasField("response"):
                                    continue

                                response = rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse()
                                response.ParseFromString(query_response.response)
                                yield response

                    except BaseException as exception:
                        # We expect to get disconnected from the placement
                        # planner from time to time, e.g., when it is
                        # being updated, but we don't want that error to
                        # propagate, we just want to retry.
                        if rbt.aio.aborted.is_grpc_retryable_exception(exception):
                            await query_backoff()
                            continue
                        raise


        def reactively(self):
            return APIKey.WeakReference._Reactively(
                service_name=rbt.aio.types.ServiceName('rbt.cloud.v1alpha1.auth.APIKeyInterface'),
                state_ref=self._state_ref,
                bearer_token=self._bearer_token,
            )

        class _Idempotently(typing.Generic[APIKey_IdempotentlyScheduleTypeVar]):

            _weak_reference: APIKey.WeakReference[APIKey_IdempotentlyScheduleTypeVar]

            def __init__(
                self,
                *,
                weak_reference: APIKey.WeakReference[APIKey_IdempotentlyScheduleTypeVar],
                idempotency: rbt.aio.idempotency.Idempotency,
            ):
                self._weak_reference = weak_reference
                self._idempotency = idempotency

            def schedule(
                self,
                *,
                when: typing.Optional[datetime | timedelta] = None,
            ) -> APIKey_IdempotentlyScheduleTypeVar:
                return self._weak_reference._schedule_type(
                    self._weak_reference._tasks,
                    when=when,
                    idempotency=self._idempotency,
                )

            async def Revoke(
                self,
                __context__: rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
                __options__: rbt.aio.call.Options = rbt.aio.call.Options(),
            ) -> rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse:
                if __options__.idempotency is not None:
                    raise RuntimeError(
                        'Found redundant idempotency in `Options`'
                    )

                __options__ = dataclasses.replace(
                    __options__,
                    idempotency_key=self._idempotency.key,
                    idempotency_alias=self._idempotency.alias,
                    generate_idempotency=self._idempotency.generate,
                )

                return await self._weak_reference.Revoke(
                    __context__,
                    __options__,
                )


        def idempotently(
            self,
            alias: typing.Optional[str] = None,
            *,
            key: typing.Optional[uuid.UUID] = None,
        ) -> APIKey.WeakReference._Idempotently[APIKey_ScheduleTypeVar]:
            return APIKey.WeakReference._Idempotently(
                weak_reference=self,
                idempotency=rbt.aio.idempotency.Idempotency(key=key, alias=alias),
            )

        def schedule(
            self,
            *,
            when: typing.Optional[datetime | timedelta] = None,
        ) -> APIKey_ScheduleTypeVar:
            return self._schedule_type(self._tasks, when=when)

        class _Schedule:

            def __init__(
                self,
                tasks: typing.Callable[[rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext], APIKeyTasksStub],
                *,
                when: typing.Optional[datetime | timedelta] = None,
                idempotency: typing.Optional[rbt.aio.idempotency.Idempotency] = None,
            ) -> None:
                self._tasks = tasks
                self._when = when
                self._idempotency = idempotency

            # APIKey callable tasks:
            @typing.overload
            async def Revoke(
                self,
                __context__: rbt.aio.contexts.TransactionContext,
                __options__: typing.Optional[rbt.aio.call.Options] = None,
            ) -> rbt.aio.tasks.TaskEffect:
                ...

            @typing.overload
            async def Revoke(
                self,
                __context__: rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
                __options__: typing.Optional[rbt.aio.call.Options] = None,
            ) -> APIKey.RevokeFuture:
                ...

            async def Revoke(
                self,
                __context__: rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
                __options__: typing.Optional[rbt.aio.call.Options] = None,
            ) -> rbt.aio.tasks.TaskEffect | APIKey.RevokeFuture:
                # TODO: mypy-protobuf declares that
                # `google.protobuf.message.Message` constructor arguments are
                # always non-None, when in reality they are optional.
                request = rbt.cloud.v1alpha1.auth.auth_pb2.RevokeRequest(
                )

                schedule: typing.Optional[datetime] = (datetime.now() + self._when) if isinstance(
                    self._when, timedelta
                ) else self._when

                metadata: typing.Optional[rbt.aio.types.GrpcMetadata] = None
                idempotency: typing.Optional[rbt.aio.idempotency.Idempotency] = self._idempotency
                bearer_token: typing.Optional[str] = None

                if __options__ is not None:
                    if __options__.idempotency:
                        if idempotency is not None:
                            raise RuntimeError(
                                'Found redundant idempotency in `Options`'
                            )
                        idempotency = __options__.idempotency
                    if __options__.metadata is not None:
                        metadata = __options__.metadata
                    if __options__.bearer_token is not None:
                        bearer_token = __options__.bearer_token

                # Add scheduling information to the metadata.
                metadata = (
                    (rbt.aio.headers.TASK_SCHEDULE,
                    schedule.isoformat() if schedule else ''),
                ) + (metadata or tuple())

                task_id = await self._tasks(
                    __context__
                ).Revoke(
                    request,
                    idempotency=idempotency,
                    metadata=metadata,
                    bearer_token=bearer_token,
                )

                if isinstance(__context__, rbt.aio.contexts.TransactionContext):
                    return rbt.aio.tasks.TaskEffect(
                        service_name=rbt.aio.types.ServiceName(
                            'rbt.cloud.v1alpha1.auth.APIKeyInterface'
                        ),
                        state_ref=__context__._state_ref,
                        method_name='Revoke',
                        request=request,
                        task_uuid=task_id.task_uuid,
                        schedule=schedule,
                    )

                return APIKey.RevokeFuture(
                    __context__,
                    task_id=task_id,
                )

            @typing.overload
            async def Authenticate(
                self,
                __context__: rbt.aio.contexts.TransactionContext,
                __options__: typing.Optional[rbt.aio.call.Options] = None,
                *,
                secret: typing.Optional[str] = None,
            ) -> rbt.aio.tasks.TaskEffect:
                ...

            @typing.overload
            async def Authenticate(
                self,
                __context__: rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
                __options__: typing.Optional[rbt.aio.call.Options] = None,
                *,
                secret: typing.Optional[str] = None,
            ) -> APIKey.AuthenticateFuture:
                ...

            async def Authenticate(
                self,
                __context__: rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
                __options__: typing.Optional[rbt.aio.call.Options] = None,
                *,
                secret: typing.Optional[str] = None,
            ) -> rbt.aio.tasks.TaskEffect | APIKey.AuthenticateFuture:
                if secret is not None and not isinstance(
                    secret,
                    str,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest': field 'secret' is not "
                        f"of required type 'str'"
                    )
                # TODO: mypy-protobuf declares that
                # `google.protobuf.message.Message` constructor arguments are
                # always non-None, when in reality they are optional.
                request = rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest(
                    secret=secret,  # type: ignore[arg-type]
                )

                schedule: typing.Optional[datetime] = (datetime.now() + self._when) if isinstance(
                    self._when, timedelta
                ) else self._when

                metadata: typing.Optional[rbt.aio.types.GrpcMetadata] = None
                idempotency: typing.Optional[rbt.aio.idempotency.Idempotency] = self._idempotency
                bearer_token: typing.Optional[str] = None

                if __options__ is not None:
                    if __options__.idempotency:
                        if idempotency is not None:
                            raise RuntimeError(
                                'Found redundant idempotency in `Options`'
                            )
                        idempotency = __options__.idempotency
                    if __options__.metadata is not None:
                        metadata = __options__.metadata
                    if __options__.bearer_token is not None:
                        bearer_token = __options__.bearer_token

                # Add scheduling information to the metadata.
                metadata = (
                    (rbt.aio.headers.TASK_SCHEDULE,
                    schedule.isoformat() if schedule else ''),
                ) + (metadata or tuple())

                task_id = await self._tasks(
                    __context__
                ).Authenticate(
                    request,
                    idempotency=idempotency,
                    metadata=metadata,
                    bearer_token=bearer_token,
                )

                if isinstance(__context__, rbt.aio.contexts.TransactionContext):
                    return rbt.aio.tasks.TaskEffect(
                        service_name=rbt.aio.types.ServiceName(
                            'rbt.cloud.v1alpha1.auth.APIKeyInterface'
                        ),
                        state_ref=__context__._state_ref,
                        method_name='Authenticate',
                        request=request,
                        task_uuid=task_id.task_uuid,
                        schedule=schedule,
                    )

                return APIKey.AuthenticateFuture(
                    __context__,
                    task_id=task_id,
                )


        class _WriterSchedule:

            def __init__(
                self,
                tasks: typing.Callable[[rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext], APIKeyTasksStub],
                *,
                when: typing.Optional[datetime | timedelta] = None,
                idempotency: typing.Optional[rbt.aio.idempotency.Idempotency] = None,
            ) -> None:
                self._tasks = tasks
                self._when = when
                self._idempotency = idempotency

            # APIKey callable tasks:
            async def Revoke(
                self,
                __context__: rbt.aio.contexts.WriterContext | rbt.aio.contexts.TransactionContext,
                __options__: typing.Optional[rbt.aio.call.Options] = None,
            ) -> rbt.aio.tasks.TaskEffect:
                # TODO: mypy-protobuf declares that
                # `google.protobuf.message.Message` constructor arguments are
                # always non-None, when in reality they are optional.
                request = rbt.cloud.v1alpha1.auth.auth_pb2.RevokeRequest(
                )
                if isinstance(__context__, rbt.aio.contexts.WriterContext):
                    return APIKeyServicerTasksStub(
                        context=__context__,
                        state_ref=__context__._state_ref,
                    ).Revoke(
                        request,
                        schedule=self._when,
                    )

                schedule: typing.Optional[datetime] = (datetime.now() + self._when) if isinstance(
                    self._when, timedelta
                ) else self._when

                metadata: typing.Optional[rbt.aio.types.GrpcMetadata] = None
                idempotency: typing.Optional[rbt.aio.idempotency.Idempotency] = self._idempotency
                bearer_token: typing.Optional[str] = None

                if __options__ is not None:
                    if __options__.idempotency is not None:
                        if idempotency is not None:
                            raise RuntimeError(
                                'Found redundant idempotency in `Options`'
                            )
                        idempotency = __options__.idempotency
                    if __options__.metadata is not None:
                        metadata = __options__.metadata
                    if __options__.bearer_token is not None:
                        bearer_token = __options__.bearer_token

                # Add scheduling information to the metadata.
                metadata = (
                    (rbt.aio.headers.TASK_SCHEDULE,
                    schedule.isoformat() if schedule else ''),
                ) + (metadata or tuple())

                task_id = await self._tasks(
                    __context__
                ).Revoke(
                    request,
                    idempotency=idempotency,
                    metadata=metadata,
                    bearer_token=bearer_token,
                )

                return rbt.aio.tasks.TaskEffect(
                    service_name=rbt.aio.types.ServiceName(
                        'rbt.cloud.v1alpha1.auth.APIKeyInterface'
                    ),
                    state_ref=__context__._state_ref,
                    method_name='Revoke',
                    request=request,
                    task_uuid=task_id.task_uuid,
                    schedule=schedule,
                )

            async def Authenticate(
                self,
                __context__: rbt.aio.contexts.WriterContext | rbt.aio.contexts.TransactionContext,
                __options__: typing.Optional[rbt.aio.call.Options] = None,
                *,
                secret: typing.Optional[str] = None,
            ) -> rbt.aio.tasks.TaskEffect:
                if secret is not None and not isinstance(
                    secret,
                    str,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest': field 'secret' is not "
                        f"of required type 'str'"
                    )
                # TODO: mypy-protobuf declares that
                # `google.protobuf.message.Message` constructor arguments are
                # always non-None, when in reality they are optional.
                request = rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest(
                    secret=secret,  # type: ignore[arg-type]
                )
                if isinstance(__context__, rbt.aio.contexts.WriterContext):
                    return APIKeyServicerTasksStub(
                        context=__context__,
                        state_ref=__context__._state_ref,
                    ).Authenticate(
                        request,
                        schedule=self._when,
                    )

                schedule: typing.Optional[datetime] = (datetime.now() + self._when) if isinstance(
                    self._when, timedelta
                ) else self._when

                metadata: typing.Optional[rbt.aio.types.GrpcMetadata] = None
                idempotency: typing.Optional[rbt.aio.idempotency.Idempotency] = self._idempotency
                bearer_token: typing.Optional[str] = None

                if __options__ is not None:
                    if __options__.idempotency is not None:
                        if idempotency is not None:
                            raise RuntimeError(
                                'Found redundant idempotency in `Options`'
                            )
                        idempotency = __options__.idempotency
                    if __options__.metadata is not None:
                        metadata = __options__.metadata
                    if __options__.bearer_token is not None:
                        bearer_token = __options__.bearer_token

                # Add scheduling information to the metadata.
                metadata = (
                    (rbt.aio.headers.TASK_SCHEDULE,
                    schedule.isoformat() if schedule else ''),
                ) + (metadata or tuple())

                task_id = await self._tasks(
                    __context__
                ).Authenticate(
                    request,
                    idempotency=idempotency,
                    metadata=metadata,
                    bearer_token=bearer_token,
                )

                return rbt.aio.tasks.TaskEffect(
                    service_name=rbt.aio.types.ServiceName(
                        'rbt.cloud.v1alpha1.auth.APIKeyInterface'
                    ),
                    state_ref=__context__._state_ref,
                    method_name='Authenticate',
                    request=request,
                    task_uuid=task_id.task_uuid,
                    schedule=schedule,
                )


        # APIKey specific methods:
        async def Revoke(
            self,
            __context__: rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
            __options__: typing.Optional[rbt.aio.call.Options] = None,
        ) -> rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse:
            # TODO: mypy-protobuf declares that
            # `google.protobuf.message.Message` constructor arguments are
            # always non-None, when in reality they are optional.
            request = rbt.cloud.v1alpha1.auth.auth_pb2.RevokeRequest(
            )
            idempotency: typing.Optional[rbt.aio.idempotency.Idempotency] = None
            metadata: typing.Optional[rbt.aio.types.GrpcMetadata] = None
            bearer_token: typing.Optional[str] = None
            if __options__ is not None:
                if __options__.idempotency is not None:
                    idempotency = __options__.idempotency
                if __options__.metadata is not None:
                    metadata = __options__.metadata
                if __options__.bearer_token is not None:
                    bearer_token = __options__.bearer_token

            return await self._writer(__context__).Revoke(
                request,
                idempotency=idempotency,
                metadata=metadata,
                bearer_token=bearer_token,
            )

        async def Authenticate(
            self,
            __context__: rbt.aio.contexts.ReaderContext | rbt.aio.contexts.WriterContext | rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
            __options__: typing.Optional[rbt.aio.call.Options] = None,
            *,
            secret: typing.Optional[str] = None,
        ) -> rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse:
            if secret is not None and not isinstance(
                secret,
                str,
            ):
                raise TypeError(
                    f"Can not construct protobuf message of type "
                    f"'rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest': field 'secret' is not "
                    f"of required type 'str'"
                )
            # TODO: mypy-protobuf declares that
            # `google.protobuf.message.Message` constructor arguments are
            # always non-None, when in reality they are optional.
            request = rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest(
                secret=secret,  # type: ignore[arg-type]
            )
            metadata: typing.Optional[rbt.aio.types.GrpcMetadata] = None
            bearer_token: typing.Optional[str] = None
            if __options__ is not None:
                if __options__.metadata is not None:
                    metadata = __options__.metadata
                if __options__.bearer_token is not None:
                    bearer_token = __options__.bearer_token
            return await self._reader(__context__).Authenticate(
                request,
                metadata=metadata,
                bearer_token=bearer_token,
            )

    @classmethod
    def lookup(
        cls,
        state_id: rbt.aio.types.StateId,
        *,
        bearer_token: typing.Optional[str] = None,
    ) -> APIKey.WeakReference[APIKey.WeakReference._Schedule]:
        return APIKey.WeakReference(
            state_id=state_id,
            schedule_type=APIKey.WeakReference._Schedule,
            bearer_token=bearer_token,
        )

    @classmethod
    async def _Create(
        cls,
        *,
        __context__: rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
        __options__: typing.Optional[rbt.aio.call.Options] = None,
        __state_id__: typing.Optional[rbt.aio.types.StateId] = None,
        __idempotency__: typing.Optional[rbt.aio.idempotency.Idempotency] = None,
        user_id: typing.Optional[str] = None,
    ) -> tuple[APIKey.WeakReference, rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse]:
        if user_id is not None and not isinstance(
            user_id,
            str,
        ):
            raise TypeError(
                f"Can not construct protobuf message of type "
                f"'rbt.cloud.v1alpha1.auth.auth_pb2.CreateRequest': field 'user_id' is not "
                f"of required type 'str'"
            )
        # TODO: mypy-protobuf declares that
        # `google.protobuf.message.Message` constructor arguments are
        # always non-None, when in reality they are optional.
        request = rbt.cloud.v1alpha1.auth.auth_pb2.CreateRequest(
            user_id=user_id,  # type: ignore[arg-type]
        )

        idempotency = __idempotency__

        metadata: typing.Optional[rbt.aio.types.GrpcMetadata] = None
        bearer_token: typing.Optional[str] = None

        if __options__ is not None:
            if __options__.idempotency is not None:
                if idempotency is not None:
                    raise RuntimeError(
                        'Found redundant idempotency in `Options`'
                    )
                idempotency = __options__.idempotency
            if __options__.metadata is not None:
                metadata = __options__.metadata
            if __options__.bearer_token is not None:
                bearer_token = __options__.bearer_token

        state_id = __state_id__

        if state_id is None:
            if idempotency is None:
                state_id = str(uuid.uuid4())
            else:
                state_id = __context__.generate_idempotent_state_id(
                    cls.__state_type_name__,
                    'Create',
                    idempotency,
                )

        reference = APIKey.lookup(
            state_id, bearer_token=bearer_token
        )
        stub = reference._writer(__context__)
        return (
            reference,
            await stub.Create(
                request,
                idempotency=idempotency,
                metadata=metadata,
                bearer_token=bearer_token,
            )
        )


    @dataclasses.dataclass(frozen=True)
    class _Construct:

        _state_id: typing.Optional[rbt.aio.types.StateId]

        def __post_init__(self) -> None:
            rbt.aio.call.validate_ascii(
                self._state_id,
                'state_id',
                MAX_ACTOR_ID_LENGTH,
                length_min=MIN_ACTOR_ID_LENGTH,
                error_type=rbt.aio.call.InvalidStateRefError,
                illegal_characters='\n\x00\x01',
            )

        async def Create(
            self,
            __context__: rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
            __options__: typing.Optional[rbt.aio.call.Options] = None,
            *,
            user_id: typing.Optional[str] = None,
        ) -> tuple[APIKey.WeakReference, rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse]:
            return await APIKey._Create(
                __context__=__context__,
                __options__=__options__,
                __state_id__=self._state_id,
                user_id=user_id,
            )


        @typing.overload
        def idempotently(self) -> APIKey._ConstructIdempotently:
            ...

        @typing.overload
        def idempotently(self, alias: str) -> APIKey._ConstructIdempotently:
            ...

        @typing.overload
        def idempotently(self, *, key: uuid.UUID) -> APIKey._ConstructIdempotently:
            ...

        def idempotently(
            self,
            alias: typing.Optional[str] = None,
            *,
            key: typing.Optional[uuid.UUID] = None,
        ) -> APIKey._ConstructIdempotently:
            return APIKey._ConstructIdempotently(
                _idempotency=rbt.aio.idempotency.Idempotency(alias=alias, key=key),
                _state_id=self._state_id,
            )

    @classmethod
    def construct(
        cls,
        *,
        id: typing.Optional[rbt.aio.types.StateId] = None,
    ) -> APIKey._Construct:
        return APIKey._Construct(_state_id=id)

    @dataclasses.dataclass(frozen=True)
    class _ConstructIdempotently:

        _idempotency: rbt.aio.idempotency.Idempotency
        _state_id: typing.Optional[rbt.aio.types.StateId] = None

        async def Create(
            self,
            __context__: rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
            __options__: typing.Optional[rbt.aio.call.Options] = None,
            *,
            user_id: typing.Optional[str] = None,
        ) -> tuple[APIKey.WeakReference, rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse]:
            return await APIKey._Create(
                __context__=__context__,
                __options__=__options__,
                __state_id__=self._state_id,
                __idempotency__=self._idempotency,
                user_id=user_id,
            )



class APIKeyServicerNodeAdaptor(APIKey.Interface):

    def __init__(self):
        # TODO: See https://github.com/reboot-dev/respect/issues/2339
        self._js_servicer_reference = self._construct_js_servicer()  # type: ignore[attr-defined]

    async def _read(self, context: rbt.aio.contexts.WorkflowContext) -> str:
        # TODO: we need to set the `application_id` context variable
        # when getting called from nodejs as it won't be set. Is there
        # a better place to set this to make sure that it isn't unset?
        # Or should we protect all calls from nodejs with this?
        with rbt.aio.internals.contextvars.use_application_id(context.application_id):

            return google.protobuf.json_format.MessageToJson(
                await super().read(context)
            )

    async def _write(
        self,
        context: rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext,
        writer: typing.Callable[[rbt.aio.contexts.WriterContext, str], typing.Awaitable[str]],
        json_options: str,
    ):
        # TODO: we need to set the `application_id` context variable
        # when getting called from nodejs as it won't be set. Is there
        # a better place to set this to make sure that it isn't unset?
        # Or should we protect all calls from nodejs with this?
        with rbt.aio.internals.contextvars.use_application_id(context.application_id):

            async def _writer(
                context: rbt.aio.contexts.WriterContext,
                state: google.protobuf.message.Message,
            ):
                state.CopyFrom(
                    google.protobuf.json_format.Parse(
                        await writer(
                            context,
                            google.protobuf.json_format.MessageToJson(state),
                        ),
                        self.__state_type__(),
                    )
                )

            options = json.loads(json_options)

            if 'idempotency' in options:
                await super().idempotently(
                    alias=options['idempotency'].get('alias'),
                    key=options['idempotency'].get('key'),
                ).write(context, _writer)
            else:
                await super().write(context, _writer)

    # APIKey specific methods:
    async def Create(
        self,
        context: rbt.aio.contexts.WriterContext,
        state: rbt.cloud.v1alpha1.auth.auth_pb2.APIKey,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.CreateRequest,
    ) -> rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse:
        json_state = google.protobuf.json_format.MessageToJson(state)
        json_request = google.protobuf.json_format.MessageToJson(request)

        aborted: asyncio.Future[None] = asyncio.Future()

        try:
            # TODO: See https://github.com/reboot-dev/respect/issues/2339
            json_result = await self._trampoline(  # type: ignore[attr-defined]
                self._js_servicer_reference,
                'writer',
                'Create',
                context,
                aborted,
                json_state,
                json_request,
            )
        except asyncio.CancelledError:
            aborted.set_result(None)
            raise
        else:
            result = json.loads(json_result)

            if 'effects' in result:
                effects = result['effects']

                state.CopyFrom(
                    google.protobuf.json_format.ParseDict(
                        effects['state'],
                        rbt.cloud.v1alpha1.auth.auth_pb2.APIKey(),
                    )
                )

                assert not hasattr(effects, 'tasks'), "JS tasks not yet implemented"

                return google.protobuf.json_format.ParseDict(
                    effects['response'],
                    rbt.cloud.v1alpha1.auth.auth_pb2.CreateResponse(),
                )
            elif 'status' in result:
                raise (
                    APIKey
                    .CreateAborted
                    .from_status(
                        google.protobuf.json_format.ParseDict(
                            result['status'],
                            status_pb2.Status(),
                        )
                    )
                )

    async def Revoke(
        self,
        context: rbt.aio.contexts.WriterContext,
        state: rbt.cloud.v1alpha1.auth.auth_pb2.APIKey,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.RevokeRequest,
    ) -> rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse:
        json_state = google.protobuf.json_format.MessageToJson(state)
        json_request = google.protobuf.json_format.MessageToJson(request)

        aborted: asyncio.Future[None] = asyncio.Future()

        try:
            # TODO: See https://github.com/reboot-dev/respect/issues/2339
            json_result = await self._trampoline(  # type: ignore[attr-defined]
                self._js_servicer_reference,
                'writer',
                'Revoke',
                context,
                aborted,
                json_state,
                json_request,
            )
        except asyncio.CancelledError:
            aborted.set_result(None)
            raise
        else:
            result = json.loads(json_result)

            if 'effects' in result:
                effects = result['effects']

                state.CopyFrom(
                    google.protobuf.json_format.ParseDict(
                        effects['state'],
                        rbt.cloud.v1alpha1.auth.auth_pb2.APIKey(),
                    )
                )

                assert not hasattr(effects, 'tasks'), "JS tasks not yet implemented"

                return google.protobuf.json_format.ParseDict(
                    effects['response'],
                    rbt.cloud.v1alpha1.auth.auth_pb2.RevokeResponse(),
                )
            elif 'status' in result:
                raise (
                    APIKey
                    .RevokeAborted
                    .from_status(
                        google.protobuf.json_format.ParseDict(
                            result['status'],
                            status_pb2.Status(),
                        )
                    )
                )

    async def Authenticate(
        self,
        context: rbt.aio.contexts.ReaderContext,
        state: rbt.cloud.v1alpha1.auth.auth_pb2.APIKey,
        request: rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateRequest,
    ) -> rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse:
        json_state = google.protobuf.json_format.MessageToJson(state)
        json_request = google.protobuf.json_format.MessageToJson(request)

        aborted: asyncio.Future[None] = asyncio.Future()

        try:
            # TODO: See https://github.com/reboot-dev/respect/issues/2339
            json_result = await self._trampoline(  # type: ignore[attr-defined]
                self._js_servicer_reference,
                'reader',
                'Authenticate',
                context,
                aborted,
                json_state,
                json_request,
            )
        except asyncio.CancelledError:
            aborted.set_result(None)
            raise
        else:
            result = json.loads(json_result)

            if 'response' in result:
                return google.protobuf.json_format.ParseDict(
                    result['response'],
                    rbt.cloud.v1alpha1.auth.auth_pb2.AuthenticateResponse(),
                )
            elif 'status' in result:
                raise (
                    APIKey
                    .AuthenticateAborted
                    .from_status(
                        google.protobuf.json_format.ParseDict(
                            result['status'],
                            status_pb2.Status(),
                        )
                    )
                )



class APIKeyWeakReferenceNodeAdaptor(APIKey.WeakReference[APIKey.WeakReference._Schedule]):

    async def _call(  # type: ignore[override]
        self,
        *,
        callable: typing.Callable[[google.protobuf.message.Message], typing.Awaitable],
        aborted_type: type[rbt.aio.aborted.Aborted],
        request_type: type[google.protobuf.message.Message],
        json_request: str,
    ) -> str:
        request = request_type()

        google.protobuf.json_format.Parse(json_request, request)

        try:
            response = await callable(request)
        except BaseException as exception:
            if isinstance(exception, aborted_type):
                return json.dumps(
                    {
                        'status': google.protobuf.json_format.MessageToDict(
                            exception.to_status()
                        )
                    }
                )
            raise
        else:
            return json.dumps(
                {
                    'response': google.protobuf.json_format.MessageToDict(
                        response
                    )
                }
            )

    async def _schedule(  # type: ignore[override]
        self,
        *,
        method: str,
        context: rbt.aio.contexts.WriterContext | rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
        schedule: datetime,
        idempotency: typing.Optional[rbt.aio.idempotency.Idempotency],
        request_type: type[google.protobuf.message.Message],
        json_request: str,
    ) -> str:
        request = request_type()

        google.protobuf.json_format.Parse(json_request, request)

        if isinstance(context, rbt.aio.contexts.WriterContext):
            task = getattr(
                APIKeyServicerTasksStub(
                    context=context,
                    state_ref=context._state_ref,
                ),
                method,
            )(request, schedule=schedule)

            return json.dumps(
                {
                    'taskId': google.protobuf.json_format.MessageToDict(
                        task.task_id
                    )
                }
            )

        # Add scheduling information to the metadata.
        metadata: rbt.aio.types.GrpcMetadata = (
            (rbt.aio.headers.TASK_SCHEDULE, schedule.isoformat()),
        )

        task_id = await getattr(super()._tasks(context), method)(
            request,
            idempotency=idempotency,
            metadata=metadata,
        )

        return json.dumps(
            {
                'taskId': google.protobuf.json_format.MessageToDict(task_id)
            }
        )

    async def _reader(  # type: ignore[override]
        self,
        method: str,
        context: rbt.aio.contexts.ReaderContext | rbt.aio.contexts.WriterContext | rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
        request_type: type[google.protobuf.message.Message],
        json_request: str,
        json_options: str,
    ) -> str:
        return await self._call(
            callable=getattr(super()._reader(context), method),
            aborted_type=getattr(
                APIKey, method + 'Aborted'
            ),
            request_type=request_type,
            json_request=json_request,
        )

    async def _writer(  # type: ignore[override]
        self,
        method: str,
        context: rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
        request_type: type[google.protobuf.message.Message],
        json_request: str,
        json_options: str,
    ) -> str:
        options = json.loads(json_options)

        idempotency: typing.Optional[rbt.aio.idempotency.Idempotency] = None

        if 'idempotency' in options:
            idempotency = rbt.aio.idempotency.Idempotency(
                alias=options['idempotency'].get('alias'),
                key=options['idempotency'].get('key'),
            )

        if 'schedule' in options:
            when = timestamp_pb2.Timestamp()
            when.FromJsonString(options['schedule']['when'])
            return await self._schedule(
                method=method,
                context=context,
                schedule=when.ToDatetime(),
                idempotency=idempotency,
                request_type=request_type,
                json_request=json_request,
            )

        callable = getattr(super()._writer(context), method)

        return await self._call(
            callable=lambda request: callable(request, idempotency=idempotency),
            aborted_type=getattr(
                APIKey, method + 'Aborted'
            ),
            request_type=request_type,
            json_request=json_request,
        )

    async def _transaction(  # type: ignore[override]
        self,
        method: str,
        context: rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
        request_type: type[google.protobuf.message.Message],
        json_request: str,
        json_options: str,
    ) -> str:
        options = json.loads(json_options)

        idempotency: typing.Optional[rbt.aio.idempotency.Idempotency] = None

        if 'idempotency' in options:
            idempotency = rbt.aio.idempotency.Idempotency(
                alias=options['idempotency'].get('alias'),
                key=options['idempotency'].get('key'),
            )

        if 'schedule' in options:
            when = timestamp_pb2.Timestamp()
            when.FromJsonString(options['schedule']['when'])
            return await self._schedule(
                method=method,
                context=context,
                schedule=when.ToDatetime(),
                idempotency=idempotency,
                request_type=request_type,
                json_request=json_request,
            )

        callable = getattr(super()._workflow(context), method)

        return await self._call(
            callable=lambda request: callable(request, idempotency=idempotency),
            aborted_type=getattr(
                APIKey, method + 'Aborted'
            ),
            request_type=request_type,
            json_request=json_request,
        )

    async def _workflow(  # type: ignore[override]
        self,
        method: str,
        context: rbt.aio.contexts.WriterContext | rbt.aio.contexts.TransactionContext | rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
        request_type: type[google.protobuf.message.Message],
        json_request: str,
        json_options: str,
    ) -> str:
        options = json.loads(json_options)

        idempotency: typing.Optional[rbt.aio.idempotency.Idempotency] = None

        if 'idempotency' in options:
            idempotency = rbt.aio.idempotency.Idempotency(
                alias=options['idempotency'].get('alias'),
                key=options['idempotency'].get('key'),
            )

        assert 'schedule' in options

        when = timestamp_pb2.Timestamp()
        when.FromJsonString(options['schedule']['when'])

        return await self._schedule(
            method=method,
            context=context,
            schedule=when.ToDatetime(),
            idempotency=idempotency,
            request_type=request_type,
            json_request=json_request,
        )

    async def _future_await(  # type: ignore[override]
        self,
        method: str,
        context: rbt.aio.contexts.WorkflowContext | rbt.aio.external.ExternalContext,
        json_task_id: str,
    ) -> str:
        future = getattr(
            APIKey, method + 'Future'
        )(
            context,
            task_id=google.protobuf.json_format.Parse(
                json_task_id,
                tasks_pb2.TaskId(),
            ),
        )

        try:
            response = await future
        except BaseException as exception:
            if isinstance(exception, rbt.aio.aborted.Aborted):
                return json.dumps(
                    {
                        'status': google.protobuf.json_format.MessageToDict(
                            exception.to_status()
                        )
                    }
                )
            raise
        else:
            return json.dumps(
                {
                    'response': google.protobuf.json_format.MessageToDict(
                        response
                    )
                }
            )

# yapf: enable
