import asyncio
import kubernetes_asyncio
import rbt.v1alpha1.istio.gateway_spec_pb2 as gateway_spec_pb2
from google.protobuf.duration_pb2 import Duration
from kubernetes_utils.kubernetes_client import AbstractEnhancedKubernetesClient
from kubernetes_utils.ownership import OwnershipInformation
from kubernetes_utils.resources.services import Port
from rbt.aio.headers import (
    APPLICATION_ID_HEADER,
    AUTHORIZATION_HEADER,
    CONSENSUS_ID_HEADER,
    IDEMPOTENCY_KEY_HEADER,
    SERVICE_NAME_HEADER,
    STATE_REF_HEADER,
    WORKFLOW_ID_HEADER,
)
from rbt.aio.types import ConsensusId, KubernetesNamespace
from rbt.controller.envoy_filter import EnvoyFilter
from rbt.controller.envoy_filter_generator import (
    FilterContext,
    generate_istio_routing_filter,
)
from rbt.controller.gateway import Gateway
from rbt.controller.placement_planner import PlacementPlanner
from rbt.controller.settings import (
    ISTIO_ALL_SIDECARS_GATEWAY_NAME,
    ISTIO_INGRESSGATEWAY_INTERNAL_PORT,
    ISTIO_INGRESSGATEWAY_LABEL_NAME,
    ISTIO_INGRESSGATEWAY_LABEL_VALUE,
    ISTIO_INGRESSGATEWAY_NAMESPACE,
    RESEMBLE_GATEWAY_NAME,
    RESEMBLE_GATEWAY_ROUTING_FILTER_NAME,
    RESEMBLE_GATEWAY_VIRTUAL_SERVICE_NAME,
    RESEMBLE_MESH_ROUTING_FILTER_NAME,
    RESEMBLE_MESH_VIRTUAL_SERVICE_NAME,
    RESEMBLE_ROUTABLE_HOSTNAME,
    USER_CONTAINER_GRPC_PORT,
    USER_CONTAINER_WEBSOCKET_PORT,
)
from rbt.controller.virtual_service import VirtualService
from rbt.v1alpha1 import placement_planner_pb2
from rbt.v1alpha1.istio.virtual_service_spec_pb2 import (
    CorsPolicy,
    Destination,
    HTTPMatchRequest,
    HTTPRoute,
    HTTPRouteDestination,
    PortSelector,
    StringMatch,
)
from respect.logging import get_logger
from typing import Coroutine, Optional

logger = get_logger(__name__)

# Browsers will first send a CORS preflight request before sending an
# HTTP request and thus we need a relatively liberal policy.
#
# TODO(benh): let users specify origins for their applications and
# then reworking this to be more conservative.
#
# NOTE: if this CORS policy changes also consider if any changes need
# to be made in rbt/routing/envoy_config.py.
CORS_POLICY = CorsPolicy(
    allow_origins=[
        StringMatch(
            # While the name 'exact' suggests that the wildcard '*' won't work,
            # it in fact does.
            exact='*'
        )
    ],
    allow_methods=['GET', 'POST', 'OPTIONS'],
    allow_headers=[
        APPLICATION_ID_HEADER,
        SERVICE_NAME_HEADER,
        STATE_REF_HEADER,
        CONSENSUS_ID_HEADER,
        IDEMPOTENCY_KEY_HEADER,
        WORKFLOW_ID_HEADER,
        'keep-alive',
        'user-agent',
        'cache-control',
        'content-type',
        'content-transfer-encoding',
        'x-accept-content-transfer-encoding',
        'x-accept-response-streaming',
        'x-user-agent',
        'grpc-timeout',
        AUTHORIZATION_HEADER,
    ],
    expose_headers=[
        'grpc-status',
        'grpc-message',
    ],
    max_age=Duration(seconds=1728000),
)


class KubernetesNetworkManager:
    """The KubernetesNetworkManager is the part of the controller that is
    responsible for managing the network configuration of the cluster. Most
    notably, it will create the `VirtualService` and `EnvoyFilter` objects
    needed to route traffic from resemble clients to the appropriate consensuses
    for serving.
    """

    def __init__(
        self,
        k8s_client: AbstractEnhancedKubernetesClient,
        placement_planner: PlacementPlanner,
        own_pod_metadata: kubernetes_asyncio.client.V1ObjectMeta,
    ):
        self._k8s_client = k8s_client
        self._own_pod_metadata = own_pod_metadata
        self._namespace = own_pod_metadata.namespace
        self._consensus_address_by_name: dict[
            ConsensusId, placement_planner_pb2.Consensus.Address] = {}
        self._consensuses: list[placement_planner_pb2.Consensus] = []

        self._placement_planner = placement_planner
        self.plan_version = -1
        self._plan_processed = asyncio.Event()
        self._stop_event = asyncio.Event()

    async def wait_for_plan_version(self, version: int):
        """Returns once a placement plan with the given version number (or
        higher) has been processed, which may be immediately."""
        while self.plan_version < version:
            await self._plan_processed.wait()

    async def run(self) -> None:
        """Syntactic sugar for `start` and `wait`."""
        await self.start()
        await self.wait()

    async def wait(self) -> None:
        """Wait until we have stopped."""
        await self._stop_event.wait()

    async def _process_plan_with_consensuses(
        self,
        plan_with_consensuses: placement_planner_pb2.ListenForPlanResponse,
    ) -> None:
        """Process a new plan-with-locations as received via callback"""

        # We rely on `time.time_ns` to provide a monotonically increasing
        # version number for the plan. As the placement planner is running
        # in the same process it is highly unlikely that we would receive
        # a plan with a lower version number, unless someone is `mock`ing
        # `time.time_ns` in a test.
        assert self.plan_version <= plan_with_consensuses.plan.version

        self.plan_version = plan_with_consensuses.plan.version
        logger.info(
            'Received new plan with '
            f'{len(plan_with_consensuses.consensuses)} consensuses'
        )

        # Update our internal view of the world to match the new consensuses.
        await self._set_consensuses(plan_with_consensuses.consensuses)

        # Indicate to anyone currently listening that a plan has been processed.
        # Then immediately clear it again, so future listeners block until the
        # next plan is processed.
        self._plan_processed.set()
        self._plan_processed.clear()
        logger.info('Waiting for next plan...')

    def is_registered_for_callbacks(self) -> bool:
        return self._process_plan_with_consensuses in self._placement_planner.plan_change_callbacks

    async def start(self) -> None:
        """Subscribe to changes in the placement planner and start
        processing callbacks."""
        logger.info('Starting KubernetesNetworkManager')

        # Subscribe to new plans.
        self._placement_planner.plan_change_callbacks.add(
            self._process_plan_with_consensuses
        )

        # Eagerly process latest plan.
        await self._process_plan_with_consensuses(
            self._placement_planner.current_plan_with_consensuses
        )

    def stop(self) -> None:
        """Unsubscribe from placement planner updates and stop processing
        callbacks."""
        logger.info('Stopping KubernetesNetworkManager')

        # Unsubscribe from new plans.
        self._placement_planner.plan_change_callbacks.remove(
            self._process_plan_with_consensuses
        )

        # Notify all waiters that we are done.
        self._stop_event.set()

    def _make_virtual_service(
        self,
        *,
        name: str,
        gateway: str,
        hosts: list[str],
    ) -> Optional[VirtualService]:
        """Produces the `VirtualService` object appropriate for the Resemble
        gateway."""
        if len(self._consensus_address_by_name) == 0:
            # If there are no consensuses, we don't need a VirtualService.
            # We also can't create an empty one; a VirtualService is required to
            # have at least one entry in its "http" list.
            return None

        # Every consensus gets two `HTTPRoute`s consisting of...
        #
        # 1a. An Istio `HTTPMatchRequest` for matching websocket
        #     traffic _and_ the header `x-reboot-consensus-id` with
        #     the value of the consensus' name.
        #
        # 1b. An Istio `HTTPRouteDestination` that describes where the
        #     request should go if it matches. In our case, that's the
        #     consensus' websocket port.
        #
        # 2a. An Istio `HTTPMatchRequest` that describes what HTTP requests
        #     should be routed to it. In our case, this looks for the header
        #     `x-reboot-consensus-id` with the value of the consensus' name.
        #
        # 2b. An Istio `HTTPRouteDestination` that describes where the request
        #     should go if it matches. In our case, that's the consensus'
        #     address.
        #
        # For example, for 2. we may have an `HTTPMatchRequest` that looks like:
        #
        #   HTTPMatchRequest(
        #       headers={
        #           'x-reboot-consensus-id': StringMatch(exact='consensus-1')
        #       }
        #   )
        #
        # And its `HTTPRouteDestination` which looks like:
        #
        #   HTTPRouteDestination(
        #       destination=Destination(
        #           host='consensus-1',
        #           port=PortSelector(number=5000)
        #       )
        #   )

        def websocket_http_routes():
            return [
                HTTPRoute(
                    name=f'websocket-route-to-{consensus_id}',
                    match=[
                        HTTPMatchRequest(
                            headers={
                                CONSENSUS_ID_HEADER:
                                    StringMatch(exact=consensus_id),
                                'upgrade':
                                    StringMatch(exact='websocket'),
                            },
                        )
                    ],
                    route=[
                        HTTPRouteDestination(
                            destination=Destination(
                                host=consensus_address.host,
                                # TODO(benh): propagate port
                                # through `ConsensusAddres`.
                                port=PortSelector(
                                    number=USER_CONTAINER_WEBSOCKET_PORT
                                )
                            ),
                        ),
                    ],
                    cors_policy=CORS_POLICY,
                ) for consensus_id, consensus_address in
                self._consensus_address_by_name.items()
            ]

        def grpc_http_routes():
            return [
                HTTPRoute(
                    name=f'grpc-route-to-{consensus_id}',
                    match=[
                        HTTPMatchRequest(
                            headers={
                                CONSENSUS_ID_HEADER:
                                    StringMatch(exact=consensus_id)
                            },
                        )
                    ],
                    route=[
                        HTTPRouteDestination(
                            destination=Destination(
                                host=consensus_address.host,
                                port=PortSelector(
                                    number=consensus_address.port
                                ),
                            ),
                        ),
                    ],
                    cors_policy=CORS_POLICY,
                ) for consensus_id, consensus_address in
                self._consensus_address_by_name.items()
            ]

        return VirtualService(
            metadata=kubernetes_asyncio.client.V1ObjectMeta(
                namespace=self._namespace,
                name=name,
            ),
            spec=VirtualService.Spec(
                gateways=[gateway],
                hosts=hosts,
                # NOTE: ordering is important here! THe websockeet
                # routes need to come first as they are more specific
                # than the gRPC routes.
                http=websocket_http_routes() + grpc_http_routes(),
            )
        )

    def _make_virtual_service_for_mesh_traffic(
        self,
        name: str,
    ) -> Optional[VirtualService]:
        """Produces the `VirtualService` object appropriate for the current set
        of consensuses.
        """
        # The VirtualService we're constructing is intended to be applied to the
        # `mesh` gateway, which means: all sidecars in the Istio mesh. It is
        # then configured to (only) act on traffic routed to resemble's "routable
        # hostname". For a discussion on why that is, see the comments in
        # `_make_virtual_service_for_gateway_traffic()`.
        return self._make_virtual_service(
            name=name,
            gateway=ISTIO_ALL_SIDECARS_GATEWAY_NAME,
            hosts=[
                # The hostname without namespace is used by traffic that
                # has a `Service` with this name in their own namespace.
                RESEMBLE_ROUTABLE_HOSTNAME,
                # The hostname with namespace is used by traffic that does
                # not have a `Service` with this name in their own
                # namespace.
                f'{RESEMBLE_ROUTABLE_HOSTNAME}.{self._namespace}',
                # It's also allowed to specify the hostname with the full
                # Kubernetes domain name.
                f'{RESEMBLE_ROUTABLE_HOSTNAME}.{self._namespace}.svc.cluster.local',
            ]
        )

    def _make_virtual_service_for_gateway_traffic(
        self,
        name: str,
    ) -> Optional[VirtualService]:
        """Produces the `VirtualService` object appropriate for the gateways.
        """
        # The VirtualService we're constructing is intended to be applied to the
        # `resemble-gateway` gateway, which runs on the Istio ingress gateways.
        # There (and ONLY there) we want to accept traffic routed to any
        # hostname, most notably the hostnames the ingress gateways may use.
        #
        # This is a separate VirtualService from the one that handles mesh
        # traffic, because it's important that ONLY the gateways handle traffic
        # addressed to the gateways. It would be incorrect for sidecars in the
        # Istio mesh to handle gateway-bound traffic: gateway traffic likely
        # comes from pods that are part of the Istio mesh but are NOT part of a
        # Resemble application. Such pods do execute all `mesh` VirtualServices
        # (whose scope is all namespaces in the Istio mesh) but DON'T run the
        # Resemble routing filters (whose scope is all namespaces in a Resemble
        # application). It is a requirement for Resemble traffic that the
        # Resemble routing filter be applied before a VirtualService is
        # executed. On pods that don't run the routing filter, we must therefore
        # avoid having a `mesh` VirtualService that acts on the traffic. We do
        # this by scoping down the `mesh` VirtualService to only act on traffic
        # that is explicitly addressed to the Resemble routable hostname (which
        # is only ever used by Resemble applications), and not on traffic routed
        # to the Resemble gateways (which is used by non-Resemble applications).
        #
        # Instead, the VirtualService for mesh traffic will ignore traffic to
        # the gateways, allowing the gateways to do routing on behalf of these
        # non-Resemble pods. The following VirtualService is the one that will
        # do that routing on the gateways.
        #
        # Once traffic is at the gateway, we accept all traffic directed to all
        # hosts. That includes hostnames we've never heard of, since the
        # operator may have configured the ingress gateway with a hostname we
        # have no knowledge of.
        return self._make_virtual_service(
            name=name,
            gateway=RESEMBLE_GATEWAY_NAME,
            hosts=["*"],
        )

    async def _write_or_delete_virtual_service(
        self, name: str, virtual_service: Optional[VirtualService]
    ):
        if virtual_service is None:
            # Delete the VirtualService if it exists.
            logger.info(f'No VirtualService needed; deleting existing {name}')
            await self._k8s_client.custom_objects.delete_by_name(
                namespace=self._namespace,
                name=name,
                object_type=VirtualService,
            )
            return
        logger.debug(f'Writing VirtualService: {virtual_service.to_dict()}')
        await self._k8s_client.custom_objects.create_or_update(virtual_service)

    async def _write_virtual_services(self):
        """Writes the latest `VirtualService`s to Kubernetes."""
        # There are two VirtualServices: one for (Istio) mesh traffic (applied
        # to Istio's `mesh` gateway which deploys it on every sidecar in every
        # namespace that's part of Istio's mesh), and one for (Resemble) gateway
        # traffic (applied to the `resemble-gateway` gateway, which deploys it
        # ONLY on ingress gateway pods).
        #
        # The `mesh` VirtualService routes traffic from Resemble actors to other
        # Resemble actors point-to-point - thanks to the routing by the sidecars
        # we don't need to route through the gateways. This saves a hop, which
        # is important for performance: otherwise the gateways would have to
        # scale up with the intra-Resemble traffic, which will be a different
        # order of magnitude than the external traffic they're required to
        # handle.
        #
        # The `resemble-gateway` VirtualService routes traffic that's arrived at
        # a Gateway. It makes the same decisions as the `mesh` VirtualService,
        # but does so at a different place.
        #
        # These VirtualServices are separate, because it would be incorrect for
        # sidecars in the Istio mesh to act on gateway-bound traffic: gateway
        # traffic likely comes from pods that are part of the Istio mesh but are
        # NOT part of a Resemble application. Such pods do execute all `mesh`
        # VirtualServices (whose scope is all namespaces in the Istio mesh) but
        # DON'T run the Resemble routing filters (since the `EnvoyFilter`s that
        # configure those are scoped to only those namespaces that are also in a
        # Resemble application). It is a requirement for Resemble traffic that
        # the Resemble routing filter be applied before their VirtualService is
        # executed. On pods that don't run the routing filter, we must therefore
        # avoid having a VirtualService that acts on the traffic. We do this by
        # scoping down the `mesh` VirtualService to only act on traffic that is
        # explicitly addressed to the Resemble "routable hostname" (which is a
        # Resemble-internal name only ever used by Resemble applications), and
        # not on traffic routed to the Resemble gateways (which is used by
        # non-Resemble applications). The traffic bound for the Resemble
        # gateways will be run through the routing filter and
        # `resemble-gateway`'s virtual service once it reaches the gateway.
        name = RESEMBLE_MESH_VIRTUAL_SERVICE_NAME
        await self._write_or_delete_virtual_service(
            name, self._make_virtual_service_for_mesh_traffic(name)
        )
        name = RESEMBLE_GATEWAY_VIRTUAL_SERVICE_NAME
        await self._write_or_delete_virtual_service(
            name, self._make_virtual_service_for_gateway_traffic(name)
        )

    async def _write_routing_filters_for_mesh_traffic(
        self, namespaces: set[KubernetesNamespace]
    ):
        """Writes the latest routing filter `EnvoyFilter` for mesh (that is:
        Consensus-to-Consensus) traffic to Kubernetes.

        The argument `namespaces` specifies the kubernetes namespaces that
        should contain a routing filter. Our routing filter in these namespaces
        will be created or updated, while routing filters previously created by
        us in other namespaces will be deleted.
        """

        # A set of envoy filters that are no longer needed; we need to remove
        # these.
        obsolete_routing_filters: list[EnvoyFilter] = []

        # Find ALL the EnvoyFilters!
        existing_routing_filters = await self._k8s_client.custom_objects.list_all(
            object_type=EnvoyFilter,
        )
        logger.debug(
            'Found %i routing filters...', len(existing_routing_filters)
        )

        # Loop over them, check if they are a) ours, b) desired.
        for existing_routing_filter in existing_routing_filters:
            name = existing_routing_filter.metadata.name
            namespace = existing_routing_filter.metadata.namespace

            if name != RESEMBLE_MESH_ROUTING_FILTER_NAME:
                # Not one of ours, skip.
                logger.debug(
                    "EnvoyFilter that is not a routing filter '%s' "
                    "encountered. Ignoring.", name
                )
                continue

            if namespace in namespaces:
                # We want to update this one, not remove it.
                logger.debug(
                    'Envoy filter in namespace %s should remain. Ignoring',
                    namespace
                )
                continue

            # Mark namespace for cleanup.
            obsolete_routing_filters.append(existing_routing_filter)

        # Remove our obsolete envoy filters.
        if len(obsolete_routing_filters) > 0:
            logger.info(
                'Removing envoy filter %s from the following namespaces: %s',
                RESEMBLE_MESH_ROUTING_FILTER_NAME,
                [
                    obsolete.metadata.namespace
                    for obsolete in obsolete_routing_filters
                ],
            )
            await asyncio.gather(
                *[
                    self._k8s_client.custom_objects.delete(obsolete)
                    for obsolete in obsolete_routing_filters
                ]
            )
        else:
            logger.info('No obsolete routing filters found')

        logger.info(
            'Writing routing filter to the following namespaces: %s',
            namespaces,
        )
        creations: list[Coroutine] = []
        for namespace in namespaces:
            # Get the `EnvoyFilter` representing our routing filter. That
            # object's `.metadata.name` is always the same, so we can safely use
            # `create_or_update` to update any previous version in-place.
            routing_filter: EnvoyFilter = generate_istio_routing_filter(
                namespace=namespace,
                name=RESEMBLE_MESH_ROUTING_FILTER_NAME,
                consensuses=self._consensuses,
                context=FilterContext.MESH,
            )
            logger.debug(f'Routing filter body: {routing_filter.to_dict()}')
            creations.append(
                self._k8s_client.custom_objects.
                create_or_update(routing_filter)
            )

        await asyncio.gather(*creations)

    async def _write_routing_filter_for_gateway_traffic(self):
        # Get the `EnvoyFilter` representing our routing filter. That object's
        # `.metadata.name` is always the same, so we can safely use
        # `create_or_update` to update any previous version in-place.
        routing_filter: EnvoyFilter = generate_istio_routing_filter(
            namespace=ISTIO_INGRESSGATEWAY_NAMESPACE,
            name=RESEMBLE_GATEWAY_ROUTING_FILTER_NAME,
            consensuses=self._consensuses,
            context=FilterContext.GATEWAY,
        )
        logger.debug(
            f'Writing gateway routing filter: {routing_filter.to_dict()}'
        )
        await self._k8s_client.custom_objects.create_or_update(routing_filter)

    async def _write_routable_service(self):
        """Writes the Kubernetes `Service` that gives us a single routable
        hostname for all services (and therefore actors) in the system.
        """
        await self._k8s_client.services.create_or_update(
            # There is one routable service for the entire Resemble system, so
            # we'll put it in the resemble system's own namespace.
            namespace=self._namespace,
            # The name of the routable service is a well-known string that all
            # clients use to contact any Resemble service.
            name=RESEMBLE_ROUTABLE_HOSTNAME,
            # The ports on which the routable service is reached is
            # universally known also.
            ports=[
                # To let this port serve gRPC traffic when there's an
                # intermediate Envoy proxy in gateway mode, this port
                # MUST be called "grpc".
                Port(port=USER_CONTAINER_GRPC_PORT, name="grpc"),
                # Port for WebSockets for browsers.
                Port(port=USER_CONTAINER_WEBSOCKET_PORT, name="websocket"),
            ],
            # We only need the routable service's hostname to exist and resolve,
            # it doesn't (shouldn't) point at any pods.
            deployment_label=None,
            # Unlike other objects created by the controller, this `Service`
            # isn't tied to a user action; it's an always-present part of the
            # system itself. Therefore, the controller itself is the owner of
            # the object.
            owner=OwnershipInformation.from_metadata(
                kind='Pod',
                api_version='v1',
                metadata=self._own_pod_metadata,
            )
        )

    async def _write_gateway(self):
        """Writes the Kubernetes `Gateway` that allows external traffic to
        reach the Resemble system.
        """
        # The Gateway object is an Istio configuration that instructs the Istio
        # gateway to accept traffic on a particular port. It also gives traffic
        # received on that port a name, which is used by VirtualServices to
        # name the traffic they'd like to be routing.
        await self._k8s_client.custom_objects.create_or_update(
            Gateway(
                metadata=kubernetes_asyncio.client.V1ObjectMeta(
                    namespace=self._namespace,
                    name=RESEMBLE_GATEWAY_NAME,
                ),
                spec=Gateway.Spec(
                    selector={
                        ISTIO_INGRESSGATEWAY_LABEL_NAME:
                            ISTIO_INGRESSGATEWAY_LABEL_VALUE
                    },
                    servers=[
                        gateway_spec_pb2.Server(
                            # Accept traffic for all hosts. We assume that if
                            # traffic can reach the gateway, we want to handle
                            # it.
                            hosts=['*'],
                            # Accept traffic on the following port.
                            #
                            # ISSUE(1529): this configures the gateway to accept
                            # (only) unencrypted traffic. Allow the user to
                            # configure their gateway to use TLS.
                            port=gateway_spec_pb2.Port(
                                number=ISTIO_INGRESSGATEWAY_INTERNAL_PORT,
                                name='grpc',
                                protocol='GRPC',
                            )
                        )
                    ]
                )
            )
        )

    async def _set_consensuses(
        self, consensuses: list[placement_planner_pb2.Consensus]
    ):
        """
        Sets the list of consensuses that the network manager knows about.
        """
        self._consensus_address_by_name = {}
        for consensus in consensuses:
            self._consensus_address_by_name[consensus.id] = consensus.address
        self._consensuses = consensuses

        # The routable service gives clients a hostname to connect to when
        # they want to talk to Resemble.
        await self._write_routable_service()

        # The Gateway allows external traffic to reach the Resemble system.
        await self._write_gateway()

        # The routing filter will select a consensus and set a consensus name
        # header based on the targeted service name and actor ID.
        # Invariant: All consensuses must have a namespace set!
        assert all([consensus.namespace for consensus in consensuses])
        namespaces: set[str] = {
            consensus.namespace
            for consensus in consensuses
            if consensus.namespace  # To satisfy mypy.
        }
        await self._write_routing_filters_for_mesh_traffic(namespaces)
        await self._write_routing_filter_for_gateway_traffic()

        # The VirtualServices direct connections to the appropriate consensus
        # based on their consensus name header.
        await self._write_virtual_services()
